// 
// Author:
//       smdn <smdn@smdn.jp>
// 
// Copyright (c) 2008-2014 smdn
// 
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
// 
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
// 
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
// THE SOFTWARE.

using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;

using Smdn.Net.Imap4.Client.Session;
using Smdn.Net.Imap4.Protocol;
using Smdn.Net.Imap4.Protocol.Client;

namespace Smdn.Net.Imap4.Client {
  public partial class ImapMailboxInfo : IImapUrl {
    public ImapClient Client {
      get { return client; }
    }

    public Uri Url {
      get { return mailbox.Url; }
    }

    public string FullName {
      get { return mailbox.Name; }
    }

    public string Name {
      get { return mailbox.LeafName; }
    }

    public string ParentMailboxName {
      get { return mailbox.SuperiorName; }
    }

    public string MailboxSeparator {
      get { return mailbox.HierarchyDelimiter; }
    }

    public long UidValidity {
      get { return mailbox.UidValidity; }
    }

    public long UnseenMessageCount {
      get { return mailbox.UnseenMessage; }
    }

    public long ExistMessageCount {
      get { return mailbox.ExistsMessage; }
    }

    public long RecentMessageCount {
      get { return mailbox.RecentMessage; }
    }

    public long NextUid {
      get { return mailbox.UidNext; }
    }

    public IImapMailboxFlagSet Flags {
      get { return mailbox.Flags; }
    }

    [CLSCompliant(false)]
    public ulong HighestModSeq {
      get { return mailbox.HighestModSeq; }
    }

    public bool IsModSequencesAvailable {
      get { return mailbox.ModificationSequences || 0 < mailbox.HighestModSeq; }
    }

    public bool Exists {
      get { return !deleted && !mailbox.Flags.Contains(ImapMailboxFlag.NonExistent); }
    }

    public bool IsUnselectable {
      get { return mailbox.IsUnselectable; }
    }

    public virtual bool IsOpen {
      get {
        if (client.OpenedMailbox == null)
          return false;
        else
          return mailbox.NameEquals(client.OpenedMailbox.mailbox);
      }
    }

    public bool IsInbox {
      get { return mailbox.IsInbox; }
    }

    public bool CanHaveChild {
      get { return !CanNotHaveChild; }
    }

    protected bool CanNotHaveChild {
      get { return mailbox.Flags.Contains(ImapMailboxFlag.NoInferiors) || string.IsNullOrEmpty(mailbox.HierarchyDelimiter); }
    }

    internal protected ImapMailbox Mailbox {
      get { return mailbox; }
    }

    internal ImapMailboxInfo(ImapClient client, ImapMailbox mailbox)
    {
      this.client = client;
      this.mailbox = mailbox;
    }

    private ImapClient client;
    private ImapMailbox mailbox;
    private bool deleted = false;

    /*
     * operations
     */
    public void Refresh()
    {
      Refresh(false);
    }

    public virtual void Refresh(bool reretrieveMailboxFlags)
    {
      if (reretrieveMailboxFlags)
        ReRetrieveMailboxFlags();

      if (!Exists || mailbox.IsUnselectable)
        return;

      ProcessResult(client.Session.Status(mailbox,
                                          client.GetAvailableStatusDataItem()));
    }

    protected void ReRetrieveMailboxFlags()
    {
      ImapMailbox[] listedMailbox;

      ProcessResult(client.Session.List(string.Empty, mailbox.Name, out listedMailbox));

      var newMailbox = listedMailbox.FirstOrDefault(m => mailbox.NameEquals(m));

      if (newMailbox == null) {
        deleted = true;
      }
      else {
        deleted = false;
        mailbox = newMailbox; // instance will be replaced if mailbox is re-created
      }
    }

    /*
     * SELECT/EXAMINE
     */
    public ImapOpenedMailboxInfo Open()
    {
      return Open(ImapClient.DefaultSelectReadOnly);
    }

    public virtual ImapOpenedMailboxInfo Open(bool asReadOnly)
    {
      return client.OpenMailbox(this, asReadOnly);
    }

    /*
     * DELETE
     */
    public void Delete()
    {
      Delete(ImapClient.DefaultSubscription);
    }

    public void Delete(bool unsubscribe)
    {
      client.CloseMailboxIfSelected(this);

      if (Exists)
        ProcessResult(client.Session.Delete(mailbox),
                      delegate(ImapTaggedStatusResponse status) {
          if (status.ResponseText.Code == ImapResponseCode.NonExistent)
            return false; // throw no exception
          else
            return true;
        });

      deleted = true;

      if (unsubscribe)
        ProcessResult(client.Session.Unsubscribe(mailbox.Name)); // mailbox is detached already
    }

    /*
     * CREATE
     */
    public void Create()
    {
      Create(ImapClient.DefaultSubscription);
    }

    public void Create(bool subscribe)
    {
      ProcessResult(client.Session.Create(mailbox.Name),
                    delegate(ImapTaggedStatusResponse status) {
        if (status.ResponseText.Code == ImapResponseCode.AlreadyExists)
          return false; // throw no exception
        else
          return true;
      });

      deleted = false;

      // remove \NonExistent flag
      if (mailbox.Flags.Contains(ImapMailboxFlag.NonExistent)) {
        var flags = new ImapMailboxFlagSet(mailbox.Flags);

        flags.Remove(ImapMailboxFlag.NonExistent);

        mailbox.Flags = flags.AsReadOnly();
      }

      if (subscribe)
        ProcessResult(client.Session.Subscribe(mailbox.Name));
    }

    public static ImapMailboxInfo Create(ImapClient client,
                                         string mailboxName)
    {
      return Create(client,
                    mailboxName,
                    ImapClient.DefaultSubscription);
    }

    public static ImapMailboxInfo Create(ImapClient client,
                                         string mailboxName,
                                         bool subscribe)
    {
      if (client == null)
        throw new ArgumentNullException("client");

      return client.CreateMailbox(mailboxName, subscribe);
    }

    /*
     * RENAME
     */
    public void MoveTo(ImapMailboxInfo destinationMailbox)
    {
      MoveTo(destinationMailbox, ImapClient.DefaultSubscription);
    }

    public void MoveTo(ImapMailboxInfo destinationMailbox,
                       bool subscribe)
    {
      if (destinationMailbox == null)
        throw new ArgumentNullException("destinationMailbox");
      if (destinationMailbox.client != this.client)
        throw new NotSupportedException("moving to mailbox of different session is not supported. use ImapClient.MoveMailbox().");

      destinationMailbox.ThrowIfCanNotHaveChild();

      MoveTo(destinationMailbox.GetFullNameOf(mailbox.LeafName),
             subscribe);
    }

    public void MoveTo(string newMailboxName)
    {
      MoveTo(newMailboxName, ImapClient.DefaultSubscription);
    }

    public void MoveTo(string newMailboxName,
                       bool subscribe)
    {
      ImapClient.GetValidatedMailboxName(newMailboxName, "newMailboxName");

      if (mailbox.NameEquals(newMailboxName))
        throw new ImapProtocolViolationException("new mailbox name must be different from current mailbox name");
      else if (ImapMailbox.IsNameInbox(newMailboxName))
        throw new ImapProtocolViolationException("can't move to INBOX");

      var isInbox = mailbox.IsInbox;
      var oldMailboxName = mailbox.Name;

      client.CloseMailboxIfSelected(this);

      Dictionary<string, ImapMailboxInfo> subscribingChildren = null;

      if (subscribe && !isInbox) {
        subscribingChildren = new Dictionary<string, ImapMailboxInfo>();

        foreach (var subscribingChild in client.GetMailboxes(mailbox, ImapMailboxListOptions.SubscribedOnly)) {
          subscribingChildren.Add(subscribingChild.FullName, subscribingChild);
        }
      }

      ProcessResult(client.Session.Rename(mailbox, newMailboxName));

      if (subscribe) {
        if (!isInbox)
          // don't UNSUBSCRIBE old INBOX
          ProcessResult(client.Session.Unsubscribe(oldMailboxName));

        if (Exists)
          ProcessResult(client.Session.Subscribe(newMailboxName));

        if (!isInbox) {
          foreach (var pair in subscribingChildren) {
            ProcessResult(client.Session.Unsubscribe(pair.Key));

            if (pair.Value.Exists)
              pair.Value.Subscribe(); // mailbox name is renamed
          }
        }
      }
    }

    /*
     * SUBSCRIBE/UNSUBSCRIBE
     */
    public void Subscribe()
    {
      Subscribe(false);
    }

    public void Subscribe(bool recursive)
    {
      ProcessResult(client.Session.Subscribe(mailbox));

      if (!recursive)
        return;

      foreach (var childMailbox in GetMailboxes(ImapMailboxListOptions.Default)) {
        ProcessResult(client.Session.Subscribe(childMailbox.mailbox));
      }
    }

    public void Unsubscribe()
    {
      Unsubscribe(false);
    }

    public void Unsubscribe(bool recursive)
    {
      ProcessResult(client.Session.Unsubscribe(mailbox));

      if (!recursive)
        return;

      foreach (var childMailbox in GetMailboxes(ImapMailboxListOptions.SubscribedOnly)) {
        ProcessResult(client.Session.Unsubscribe(childMailbox.mailbox));
      }
    }

    /*
     * LIST/LSUB/CREATE
     */
    public IEnumerable<ImapMailboxInfo> GetMailboxes()
    {
      return GetMailboxes(ImapMailboxListOptions.Default);
    }

    public IEnumerable<ImapMailboxInfo> GetMailboxes(ImapMailboxListOptions options)
    {
      return client.GetMailboxes(mailbox, options);
    }

    public ImapMailboxInfo GetParent()
    {
      return GetParent(ImapMailboxListOptions.Default);
    }

    public ImapMailboxInfo GetParent(ImapMailboxListOptions options)
    {
      if (mailbox.SuperiorName.Length == 0)
        throw new ImapProtocolViolationException(string.Format("mailbox '{0}' is top-level mailbox", FullName));

      return client.GetMailbox(mailbox.SuperiorName, options);
    }

    public ImapMailboxInfo GetOrCreateParent()
    {
      return GetOrCreateParent(ImapClient.DefaultSubscription,
                               ImapMailboxListOptions.Default);
    }

    public ImapMailboxInfo GetOrCreateParent(ImapMailboxListOptions options)
    {
      return GetOrCreateParent(ImapClient.DefaultSubscription,
                               options);
    }

    public ImapMailboxInfo GetOrCreateParent(bool subscribeIfCreated)
    {
      return GetOrCreateParent(subscribeIfCreated,
                               ImapMailboxListOptions.Default);
    }

    public ImapMailboxInfo GetOrCreateParent(bool subscribeIfCreated,
                                             ImapMailboxListOptions options)
    {
      if (mailbox.SuperiorName.Length == 0)
        throw new ImapProtocolViolationException(string.Format("mailbox '{0}' is top-level mailbox", FullName));

      return client.GetOrCreateMailbox(mailbox.SuperiorName,
                                       subscribeIfCreated,
                                       options);
    }

    /// <exception cref="ImapProtocolViolationException">Throws if <see cref="CanHaveChild"/> is false.</exception>
    public string GetFullNameOf(string inferiorName)
    {
      ImapClient.GetValidatedMailboxName(inferiorName, "inferiorName");

      ThrowIfCanNotHaveChild();

      return mailbox.GetChildName(inferiorName);
    }

    /// <exception cref="ImapProtocolViolationException">Throws if <see cref="CanHaveChild"/> is false.</exception>
    public string GetFullNameOf(string inferiorName1, string inferiorName2)
    {
      return GetFullNameOf((IEnumerable<string>)new string[] {inferiorName1, inferiorName2});
    }

    /// <exception cref="ImapProtocolViolationException">Throws if <see cref="CanHaveChild"/> is false.</exception>
    public string GetFullNameOf(string inferiorName1, string inferiorName2, string inferiorName3)
    {
      return GetFullNameOf((IEnumerable<string>)new string[] {inferiorName1, inferiorName2, inferiorName3});
    }

    /// <exception cref="ImapProtocolViolationException">Throws if <see cref="CanHaveChild"/> is false.</exception>
    public string GetFullNameOf(params string[] inferiorNames)
    {
      return GetFullNameOf((IEnumerable<string>)inferiorNames);
    }

    /// <exception cref="ImapProtocolViolationException">Throws if <see cref="CanHaveChild"/> is false.</exception>
    public string GetFullNameOf(IEnumerable<string> inferiorNames)
    {
      if (inferiorNames == null)
        throw new ArgumentNullException("inferiorNames");

      ThrowIfCanNotHaveChild();

      var builder = new StringBuilder(128);

      builder.Append(mailbox.Name);

      foreach (var inferiorName in inferiorNames) {
        builder.Append(mailbox.HierarchyDelimiter);
        builder.Append(ImapClient.GetValidatedMailboxName(inferiorName, "inferiorNames"));
      }

      return builder.ToString();
    }

    /// <exception cref="ImapProtocolViolationException">Throws if <see cref="CanHaveChild"/> is false.</exception>
    public string[] SplitMailboxHierarchy(string mailboxName)
    {
      ImapClient.GetValidatedMailboxName(mailboxName, "mailboxName");

      ThrowIfCanNotHaveChild();

      return mailboxName.Split(new[] {mailbox.HierarchyDelimiter}, StringSplitOptions.None);
    }

    /// <exception cref="ImapProtocolViolationException">Throws if <see cref="CanHaveChild"/> is false.</exception>
    public ImapMailboxInfo CreateChild(string name)
    {
      return CreateChild(name, ImapClient.DefaultSubscription);
    }

    /// <exception cref="ImapProtocolViolationException">Throws if <see cref="CanHaveChild"/> is false.</exception>
    public ImapMailboxInfo CreateChild(string name,
                                       bool subscribe)
    {
      return client.CreateMailbox(GetFullNameOf(name), subscribe);
    }

    /// <exception cref="ImapProtocolViolationException">Throws if <see cref="CanHaveChild"/> is false.</exception>
    public ImapMailboxInfo GetChild(string name)
    {
      return GetChild(name, ImapMailboxListOptions.Default);
    }

    /// <exception cref="ImapProtocolViolationException">Throws if <see cref="CanHaveChild"/> is false.</exception>
    public ImapMailboxInfo GetChild(string name,
                                    ImapMailboxListOptions options)
    {
      return client.GetMailbox(GetFullNameOf(name), options);
    }

    /// <exception cref="ImapProtocolViolationException">Throws if <see cref="CanHaveChild"/> is false.</exception>
    public ImapMailboxInfo GetOrCreateChild(string name)
    {
      return GetOrCreateChild(name,
                              ImapClient.DefaultSubscription,
                              ImapMailboxListOptions.Default);
    }

    /// <exception cref="ImapProtocolViolationException">Throws if <see cref="CanHaveChild"/> is false.</exception>
    public ImapMailboxInfo GetOrCreateChild(string name,
                                            ImapMailboxListOptions options)
    {
      return GetOrCreateChild(name,
                              ImapClient.DefaultSubscription,
                              options);
    }

    /// <exception cref="ImapProtocolViolationException">Throws if <see cref="CanHaveChild"/> is false.</exception>
    public ImapMailboxInfo GetOrCreateChild(string name,
                                            bool subscribeIfCreated)
    {
      return GetOrCreateChild(name,
                              subscribeIfCreated,
                              ImapMailboxListOptions.Default);
    }

    /// <exception cref="ImapProtocolViolationException">Throws if <see cref="CanHaveChild"/> is false.</exception>
    public ImapMailboxInfo GetOrCreateChild(string name,
                                            bool subscribeIfCreated,
                                            ImapMailboxListOptions options)
    {
      var mailbox = client.GetMailboxNoException(GetFullNameOf(name), options);

      if (mailbox == null)
        return CreateChild(name, subscribeIfCreated);
      else
        return mailbox;
    }

    /*
     * GETQUOTAROOT
     */
    public IEnumerable<ImapQuota> GetQuota()
    {
      if (!client.IsCapabilityAvailable(ImapCapability.Quota))
        yield break;

      IDictionary<string, ImapQuota[]> quotaRoots;

      ProcessResult(client.Session.GetQuotaRoot(Mailbox, out quotaRoots));

      foreach (var pair in quotaRoots) {
        foreach (var quota in pair.Value) {
          yield return quota;
        }
      }
    }

    internal protected ImapCommandResult ProcessResult(ImapCommandResult result)
    {
      return ProcessResult(result, null);
    }

    internal protected virtual ImapCommandResult ProcessResult(ImapCommandResult result,
                                                               Func<ImapTaggedStatusResponse, bool> throwIfErrorResponse)
    {
      return client.ProcessResult(result, throwIfErrorResponse);
    }

    internal void ThrowIfCanNotHaveChild()
    {
      if (CanNotHaveChild)
        throw new ImapProtocolViolationException(string.Format("It is not possible for any child levels of hierarchy to exist under this mailbox. (mailbox name: '{0}')", FullName));
    }

    internal void ThrowIfNonExistentOrUnselectable()
    {
      if (!Exists)
        throw new ImapProtocolViolationException(string.Format("mailbox '{0}' is not existent", FullName));
      if (mailbox.IsUnselectable)
        throw new ImapProtocolViolationException(string.Format("mailbox '{0}' is unselectable", FullName));
    }

    /// <summary>Infrastructure. It is not intended to be used directly from your code.</summary>
    void IImapUrl.SetBaseUrl(ImapUriBuilder baseUrl)
    {
      throw new NotSupportedException();
    }

    public override string ToString()
    {
      return string.Format("{{ImapMailboxInfo: Authority='{0}', FullName='{1}', UidValidity={2}, NextUid={3}}}",
                           ImapStyleUriParser.GetStrongAuthority(Url),
                           FullName,
                           UidValidity,
                           NextUid);
    }
  }
}
