// 
// Author:
//       smdn <smdn@smdn.jp>
// 
// Copyright (c) 2008-2014 smdn
// 
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
// 
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
// 
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
// THE SOFTWARE.

using System;

using Smdn.Net.Imap4.Client.Session;
using Smdn.Net.Imap4.Protocol;
using Smdn.Net.Imap4.Protocol.Client;

namespace Smdn.Net.Imap4.Client {
  public abstract class ImapMessageInfoBase {
    public ImapClient Client {
      get { return client; }
    }

    public ImapOpenedMailboxInfo Mailbox {
      get { CheckAttachedToSelectedMailbox(); return mailbox; }
    }

    public bool IsAvailable {
      get { return mailbox != null && client.OpenedMailbox == mailbox; }
    }

    public long UidValidity {
      get; protected set;
    }

    private readonly ImapClient client;
    private ImapOpenedMailboxInfo mailbox;

    protected ImapMessageInfoBase(ImapOpenedMailboxInfo mailbox)
    {
      if (mailbox == null)
        throw new ArgumentNullException("mailbox");

      this.client = mailbox.Client;
      this.mailbox = mailbox;
      this.UidValidity = mailbox.UidValidity; // save current UIDVALIDITY
    }

    protected abstract ImapSequenceSet GetSequenceOrUidSet();

    protected void CheckAttachedToSelectedMailbox()
    {
      if (!IsAvailable)
        throw new ImapUnavailableException();
    }

    protected void CheckMessageNumberAvailable(ImapSequenceSet sequenceOrUidSet)
    {
      CheckAttachedToSelectedMailbox();

      mailbox.CheckUidValidity(mailbox.UidValidity, sequenceOrUidSet);
    }

    internal void DetachFromMailbox()
    {
      mailbox = null;
    }

    /*
     * operations
     */
    public void AddFlags(ImapMessageFlag flag,
                         params ImapMessageFlag[] flags)
    {
      StoreCore(ImapStoreDataItem.AddFlags(flag, flags));
    }

    public void AddKeywords(string keyword,
                            params string[] keywords)
    {
      StoreCore(ImapStoreDataItem.AddFlags(keyword, keywords));
    }

    public void AddFlags(IImapMessageFlagSet flagsAndKeywords)
    {
      StoreCore(ImapStoreDataItem.AddFlags(flagsAndKeywords));
    }

    public void RemoveFlags(ImapMessageFlag flag,
                            params ImapMessageFlag[] flags)
    {
      StoreCore(ImapStoreDataItem.RemoveFlags(flag, flags));
    }

    public void RemoveKeywords(string keyword,
                               params string[] keywords)
    {
      StoreCore(ImapStoreDataItem.RemoveFlags(keyword, keywords));
    }

    public void RemoveFlags(IImapMessageFlagSet flagsAndKeywords)
    {
      StoreCore(ImapStoreDataItem.RemoveFlags(flagsAndKeywords));
    }

    public void SetFlags(ImapMessageFlag flag,
                          params ImapMessageFlag[] flags)
    {
      StoreCore(ImapStoreDataItem.ReplaceFlags(flag, flags));
    }

    public void SetKeywords(string keyword,
                            params string[] keywords)
    {
      StoreCore(ImapStoreDataItem.ReplaceFlags(keyword, keywords));
    }

    public void SetFlags(IImapMessageFlagSet flagsAndKeywords)
    {
      StoreCore(ImapStoreDataItem.ReplaceFlags(flagsAndKeywords));
    }

    [Obsolete("use SetFlags() instead", false)]
    public void ReplaceFlags(ImapMessageFlag flag,
                             params ImapMessageFlag[] flags)
    {
      StoreCore(ImapStoreDataItem.ReplaceFlags(flag, flags));
    }

    [Obsolete("use SetKeywords() instead", false)]
    public void ReplaceKeywords(string keyword,
                                params string[] keywords)
    {
      StoreCore(ImapStoreDataItem.ReplaceFlags(keyword, keywords));
    }

    [Obsolete("use SetFlags() instead", false)]
    public void ReplaceFlags(IImapMessageFlagSet flagsAndKeywords)
    {
      StoreCore(ImapStoreDataItem.ReplaceFlags(flagsAndKeywords));
    }

    public virtual void Store(ImapStoreDataItem storeDataItem)
    {
      if (storeDataItem == null)
        throw new ArgumentNullException("storeDataItem");

      StoreCore(storeDataItem);
    }

    private ImapSequenceSet StoreCore(ImapStoreDataItem storeDataItem)
    {
      var sequenceOrUidSet = GetSequenceOrUidSet();

      StoreCore(sequenceOrUidSet, storeDataItem);

      return sequenceOrUidSet;
    }

    private void StoreCore(ImapSequenceSet sequenceOrUidSet,
                           ImapStoreDataItem storeDataItem)
    {
      if (sequenceOrUidSet.IsEmpty)
        return; // do nothing

      CheckMessageNumberAvailable(sequenceOrUidSet);

      mailbox.ProcessResult(client.Session.Store(sequenceOrUidSet, storeDataItem));
    }

    private static ImapStoreDataItem markAsDeletedStoreDataItem = ImapStoreDataItem.AddFlags(ImapMessageFlag.Deleted);

    public void MarkAsDeleted()
    {
      StoreCore(markAsDeletedStoreDataItem);
    }

    private static ImapStoreDataItem markAsSeenStoreDataItem = ImapStoreDataItem.AddFlags(ImapMessageFlag.Seen);

    public void MarkAsSeen()
    {
      StoreCore(markAsSeenStoreDataItem);
    }

    public void Delete()
    {
      var sequenceOrUidSet = StoreCore(markAsDeletedStoreDataItem);

      if (sequenceOrUidSet.IsEmpty)
        return; // do nothing

      if (sequenceOrUidSet.IsUidSet && client.IsCapabilityAvailable(ImapCapability.UidPlus))
        mailbox.ProcessResult(client.Session.UidExpunge(sequenceOrUidSet));
      else
        mailbox.ProcessResult(client.Session.Expunge());

      if (IsAvailable)
        mailbox.Detach(this);
    }

    public void CopyTo(ImapMailboxInfo destinationMailbox)
    {
      if (destinationMailbox == null)
        throw new ArgumentNullException("destinationMailbox");

      CopyToMoveToCore(ImapSession.CopyMoveCommand.Copy, destinationMailbox);
    }

    public void MoveTo(ImapMailboxInfo destinationMailbox)
    {
      if (destinationMailbox == null)
        throw new ArgumentNullException("destinationMailbox");

      CheckAttachedToSelectedMailbox();

      if (mailbox.Mailbox.NameEquals(destinationMailbox.Mailbox))
        return; // do nothing

      CopyToMoveToCore(ImapSession.CopyMoveCommand.Move, destinationMailbox);
    }

    private void CopyToMoveToCore(ImapSession.CopyMoveCommand command,
                                  ImapMailboxInfo destinationMailbox)
    {
      destinationMailbox.ThrowIfNonExistentOrUnselectable();

      var sequenceOrUidSet = GetSequenceOrUidSet();

      if (sequenceOrUidSet.IsEmpty)
        return; // do nothing

      CheckMessageNumberAvailable(sequenceOrUidSet);

      var move = command == ImapSession.CopyMoveCommand.Move;
      var markAsDelete = false;

      if (destinationMailbox.Client == this.client) {
        if (move && !client.IsCapabilityAvailable(ImapCapability.Move)) {
          // use COPY+STORE instead of MOVE
          command = ImapSession.CopyMoveCommand.Copy;
          markAsDelete = true;
        }

        mailbox.ProcessResult(client.Session.Copy(command, sequenceOrUidSet, destinationMailbox.Mailbox));
      }
      else {
        if (move)
          markAsDelete = true;

        var appendMessage = this as ImapMessageInfo;

        if (appendMessage == null) {
          var appendMessages = new ImapMessageInfoList(mailbox,
                                                       ImapMessageFetchAttributeOptions.None,
                                                       sequenceOrUidSet);

          foreach (var message in appendMessages) {
            destinationMailbox.AppendMessageInternal(message);
          }
        }
        else {
          destinationMailbox.AppendMessageInternal(appendMessage);
        }
      }

      if (markAsDelete)
        StoreCore(sequenceOrUidSet, markAsDeletedStoreDataItem);

      if (move && IsAvailable)
        mailbox.Detach(this);
    }

    public void CopyTo(string destinationMailboxName)
    {
      CopyToMoveToCore(ImapSession.CopyMoveCommand.Copy,
                       destinationMailboxName,
                       false);
    }

    /// <returns>The mailbox if created, otherwise null.</returns>
    public ImapMailboxInfo CopyTo(string destinationMailboxName,
                                  bool tryCreate)
    {
      return CopyToMoveToCore(ImapSession.CopyMoveCommand.Copy,
                              destinationMailboxName,
                              tryCreate);
    }

    public void MoveTo(string destinationMailboxName)
    {
      CheckAttachedToSelectedMailbox();

      if (mailbox.Mailbox.NameEquals(destinationMailboxName))
        return; // do nothing

      CopyToMoveToCore(ImapSession.CopyMoveCommand.Move,
                       destinationMailboxName,
                       false);
    }

    /// <returns>The mailbox if created, otherwise null.</returns>
    public ImapMailboxInfo MoveTo(string destinationMailboxName,
                                  bool tryCreate)
    {
      CheckAttachedToSelectedMailbox();

      if (mailbox.Mailbox.NameEquals(destinationMailboxName))
        return null; // do nothing

      return CopyToMoveToCore(ImapSession.CopyMoveCommand.Move,
                              destinationMailboxName,
                              tryCreate);
    }

    private ImapMailboxInfo CopyToMoveToCore(ImapSession.CopyMoveCommand command,
                                             string destinationMailboxName,
                                             bool tryCreate)
    {
      ImapClient.GetValidatedMailboxName(destinationMailboxName, "destinationMailboxName");

      var sequenceOrUidSet = GetSequenceOrUidSet();

      if (sequenceOrUidSet.IsEmpty)
        return null; // do nothing

      var move = (command == ImapSession.CopyMoveCommand.Move);
      var markAsDelete = false;

      if (move && !client.IsCapabilityAvailable(ImapCapability.Move)) {
        // use COPY+STORE instead of MOVE
        command = ImapSession.CopyMoveCommand.Copy;
        markAsDelete = true;
      }

      CheckMessageNumberAvailable(sequenceOrUidSet);

      ImapMailbox createdMailbox = null;

      if (tryCreate) {
        mailbox.ProcessResult(client.Session.Copy(command,
                                                  sequenceOrUidSet,
                                                  destinationMailboxName,
                                                  out createdMailbox));
      }
      else {
        mailbox.ProcessResult(client.Session.Copy(command,
                                                  sequenceOrUidSet,
                                                  destinationMailboxName),
                              delegate(ImapTaggedStatusResponse status) {
                                if (status.ResponseText.Code == ImapResponseCode.TryCreate)
                                  throw new ImapMailboxNotFoundException(destinationMailboxName);
                                else
                                  return true;
                              });
      }

      var createdMailboxInfo = (createdMailbox == null)
        ? null
        : client.CreateFlagsRetrieved(createdMailbox);

      if (markAsDelete)
        StoreCore(sequenceOrUidSet, markAsDeletedStoreDataItem);

      if (move && IsAvailable)
        mailbox.Detach(this);

      return createdMailboxInfo;
    }
  }
}
