using System;
using System.Linq;
using System.Text;
using NUnit.Framework;

namespace Smdn.Net.Imap4.Client {
  [TestFixture]
  public class ImapMessageInfoBaseTests {
    private class MessageSet : ImapMessageInfoBase {
      public MessageSet(ImapOpenedMailboxInfo mailbox, ImapSequenceSet sequenceOrUidSet)
        : base(mailbox)
      {
        this.sequenceOrUidSet = sequenceOrUidSet;
      }

      protected override ImapSequenceSet GetSequenceOrUidSet()
      {
        return sequenceOrUidSet;
      }

      private ImapSequenceSet sequenceOrUidSet;
    }

    private void TestMessage(int existMesasgeCount, ImapSequenceSet sequenceOrUidSet, Action<ImapPseudoServer, ImapMessageInfoBase, ImapMessageInfo[]> action)
    {
      TestMessage(existMesasgeCount, null, sequenceOrUidSet, action);
    }

    private void TestMessage(int existMessageCount, ImapCapability[] capabilities, ImapSequenceSet sequenceOrUidSet, Action<ImapPseudoServer, ImapMessageInfoBase, ImapMessageInfo[]> action)
    {
      TestUtils.TestAuthenticated(capabilities, delegate(ImapPseudoServer server, ImapClient client) {
        // LIST
        server.EnqueueTaggedResponse("* LIST () \".\" INBOX\r\n" +
                                     "$tag OK done\r\n");
        // SELECT
        server.EnqueueTaggedResponse(string.Format("* EXISTS {0}\r\n", existMessageCount) +
                                     "* OK [UIDVALIDITY 1]\r\n" +
                                     "$tag OK done\r\n");

        using (var inbox = client.OpenInbox()) {
          server.DequeueRequest(); // LIST
          server.DequeueRequest(); // SELECT

          var messages = new ImapMessageInfo[0];

          if (0 < existMessageCount) {
            // NOOP
            server.EnqueueTaggedResponse("$tag OK done\r\n");
            // FETCH
            var fetchResp = new StringBuilder();

            for (var seq = 1; seq <= existMessageCount; seq++) {
              fetchResp.AppendFormat("* FETCH {0} (UID {0})\r\n", seq);
            }

            fetchResp.Append("$tag OK done\r\n");

            server.EnqueueTaggedResponse(fetchResp.ToString());

            var msgs = inbox.GetMessages();

            messages = msgs.ToArray();

            server.DequeueRequest(); // NOOP
            server.DequeueRequest(); // FETCH

            Assert.AreSame(client, msgs.Client);
            Assert.AreSame(inbox, msgs.Mailbox);
            Assert.IsTrue(msgs.IsAvailable);
          }

          try {
            var messageSet = new MessageSet(inbox, sequenceOrUidSet);

            Assert.AreSame(client, messageSet.Client);
            Assert.AreSame(inbox, messageSet.Mailbox);
            Assert.IsTrue(messageSet.IsAvailable);
            Assert.AreEqual(1L, messageSet.UidValidity);

            action(server, messageSet, messages);
          }
          finally {
            if (inbox.IsOpen)
              // CLOSE
              server.EnqueueTaggedResponse("$tag OK done\r\n");
          }
        }
      });
    }

    private ImapSequenceSet CreateEmptySet()
    {
      return ImapSequenceSet.CreateSet(new long[0]);
    }

    [Test]
    public void TestAddFlags()
    {
      TestMessage(1, ImapSequenceSet.CreateUidSet(1L),
                  delegate(ImapPseudoServer server, ImapMessageInfoBase messageSet, ImapMessageInfo[] messages) {
        // UID STORE
        server.EnqueueTaggedResponse("* FETCH 1 (FLAGS (\\Seen \\Draft))\r\n" +
                                     "$tag OK done\r\n");

        messageSet.AddFlags(ImapMessageFlag.Draft);

        Assert.That(server.DequeueRequest(), Is.StringEnding("UID STORE 1 +FLAGS (\\Draft)\r\n"));

        Assert.IsNotNull(messages[0].Flags);
        Assert.AreEqual(2, messages[0].Flags.Count);
        Assert.IsTrue(messages[0].Flags.Contains(ImapMessageFlag.Seen));
        Assert.IsTrue(messages[0].Flags.Contains(ImapMessageFlag.Draft));
        Assert.IsTrue(messages[0].IsSeen);
        Assert.IsTrue(messages[0].IsDraft);
      });
    }

    [Test]
    public void TestRemoveFlags()
    {
      TestMessage(1, ImapSequenceSet.CreateUidSet(1L),
                  delegate(ImapPseudoServer server, ImapMessageInfoBase messageSet, ImapMessageInfo[] messages) {
        // UID STORE
        server.EnqueueTaggedResponse("* FETCH 1 (FLAGS (\\Seen))\r\n" +
                                     "$tag OK done\r\n");

        messageSet.RemoveFlags(ImapMessageFlag.Draft);

        Assert.That(server.DequeueRequest(), Is.StringEnding("UID STORE 1 -FLAGS (\\Draft)\r\n"));

        Assert.IsNotNull(messages[0].Flags);
        Assert.AreEqual(1, messages[0].Flags.Count);
        Assert.IsTrue(messages[0].Flags.Contains(ImapMessageFlag.Seen));
        Assert.IsTrue(messages[0].IsSeen);
      });
    }

    [Test]
    public void TestSetFlags()
    {
      TestMessage(1, ImapSequenceSet.CreateUidSet(1L),
                  delegate(ImapPseudoServer server, ImapMessageInfoBase messageSet, ImapMessageInfo[] messages) {
        // UID STORE
        server.EnqueueTaggedResponse("* FETCH 1 (FLAGS (\\Draft \\Flagged))\r\n" +
                                     "$tag OK done\r\n");

        messageSet.SetFlags(ImapMessageFlag.Draft, ImapMessageFlag.Flagged);

        Assert.That(server.DequeueRequest(), Is.StringEnding("UID STORE 1 FLAGS (\\Draft \\Flagged)\r\n"));

        Assert.IsNotNull(messages[0].Flags);
        Assert.AreEqual(2, messages[0].Flags.Count);
        Assert.IsTrue(messages[0].Flags.Contains(ImapMessageFlag.Draft));
        Assert.IsTrue(messages[0].Flags.Contains(ImapMessageFlag.Flagged));
        Assert.IsTrue(messages[0].IsDraft);
        Assert.IsTrue(messages[0].IsFlagged);
      });
    }

    [Test]
    public void TestAddFlagsNonApplicableFlagsRemoved()
    {
      TestMessage(1, ImapSequenceSet.CreateUidSet(1L),
                  delegate(ImapPseudoServer server, ImapMessageInfoBase messageSet, ImapMessageInfo[] messages) {
        // UID STORE
        server.EnqueueTaggedResponse("* FETCH 1 (FLAGS (\\Draft))\r\n" +
                                     "$tag OK done\r\n");

        messageSet.AddFlags(ImapMessageFlag.Draft, ImapMessageFlag.Recent);

        Assert.That(server.DequeueRequest(), Is.StringEnding("UID STORE 1 +FLAGS (\\Draft)\r\n"));

        Assert.IsNotNull(messages[0].Flags);
        Assert.AreEqual(1, messages[0].Flags.Count);
        Assert.IsTrue(messages[0].Flags.Contains(ImapMessageFlag.Draft));
        Assert.IsFalse(messages[0].Flags.Contains(ImapMessageFlag.Recent));
      });
    }

    [Test]
    public void TestRemoveFlagsNonApplicableFlagsRemoveds()
    {
      TestMessage(1, ImapSequenceSet.CreateUidSet(1L),
                  delegate(ImapPseudoServer server, ImapMessageInfoBase messageSet, ImapMessageInfo[] messages) {
        // UID STORE
        server.EnqueueTaggedResponse("* FETCH 1 (FLAGS (\\Draft \\Recent))\r\n" +
                                     "$tag OK done\r\n");

        messageSet.RemoveFlags(ImapMessageFlag.Draft, ImapMessageFlag.Recent);

        Assert.That(server.DequeueRequest(), Is.StringEnding("UID STORE 1 -FLAGS (\\Draft)\r\n"));

        Assert.IsNotNull(messages[0].Flags);
        Assert.AreEqual(2, messages[0].Flags.Count);
        Assert.IsTrue(messages[0].Flags.Contains(ImapMessageFlag.Draft));
        Assert.IsTrue(messages[0].Flags.Contains(ImapMessageFlag.Recent));
      });
    }

    [Test]
    public void TestSetFlagsNonApplicableFlagsRemoved()
    {
      TestMessage(1, ImapSequenceSet.CreateUidSet(1L),
                  delegate(ImapPseudoServer server, ImapMessageInfoBase messageSet, ImapMessageInfo[] messages) {
        // UID STORE
        server.EnqueueTaggedResponse("* FETCH 1 (FLAGS (\\Draft))\r\n" +
                                     "$tag OK done\r\n");

        messageSet.SetFlags(ImapMessageFlag.Draft, ImapMessageFlag.Recent);

        Assert.That(server.DequeueRequest(), Is.StringEnding("UID STORE 1 FLAGS (\\Draft)\r\n"));

        Assert.IsNotNull(messages[0].Flags);
        Assert.AreEqual(1, messages[0].Flags.Count);
        Assert.IsTrue(messages[0].Flags.Contains(ImapMessageFlag.Draft));
        Assert.IsFalse(messages[0].Flags.Contains(ImapMessageFlag.Recent));
      });
    }

    [Test]
    public void TestAddFlagsContainsNoApplicableFlags()
    {
      TestMessage(1, ImapSequenceSet.CreateUidSet(1L),
                  delegate(ImapPseudoServer server, ImapMessageInfoBase messageSet, ImapMessageInfo[] messages) {
        Assert.Throws<ArgumentException>(() => messageSet.AddFlags(ImapMessageFlag.Recent));
      });
    }

    [Test]
    public void TestRemoveFlagsContainsNoApplicableFlags()
    {
      TestMessage(1, ImapSequenceSet.CreateUidSet(1L),
                  delegate(ImapPseudoServer server, ImapMessageInfoBase messageSet, ImapMessageInfo[] messages) {
        Assert.Throws<ArgumentException>(() => messageSet.RemoveFlags(ImapMessageFlag.Recent));
      });
    }

    [Test]
    public void TestSetFlagsContainsNoApplicableFlags()
    {
      TestMessage(1, ImapSequenceSet.CreateUidSet(1L),
                  delegate(ImapPseudoServer server, ImapMessageInfoBase messageSet, ImapMessageInfo[] messages) {
        Assert.Throws<ArgumentException>(() => messageSet.SetFlags(ImapMessageFlag.Recent));
      });
    }

    [Test]
    public void TestStoreEmptySet()
    {
      TestMessage(0, CreateEmptySet(),
                  delegate(ImapPseudoServer server, ImapMessageInfoBase messageSet, ImapMessageInfo[] messages) {
        // do nothing
        messageSet.Store(ImapStoreDataItem.AddFlags(ImapMessageFlag.Seen));
      });
    }

    [Test]
    public void TestStoreUidSingle()
    {
      TestMessage(1, ImapSequenceSet.CreateUidSet(1L),
                  delegate(ImapPseudoServer server, ImapMessageInfoBase messageSet, ImapMessageInfo[] messages) {
        // UID STORE
        server.EnqueueTaggedResponse("* FETCH 1 (FLAGS (\\Seen))\r\n" +
                                     "$tag OK done\r\n");

        messageSet.Store(ImapStoreDataItem.AddFlags(ImapMessageFlag.Seen));

        Assert.That(server.DequeueRequest(), Is.StringEnding("UID STORE 1 +FLAGS (\\Seen)\r\n"));

        Assert.IsNotNull(messages[0].Flags);
        Assert.AreEqual(1, messages[0].Flags.Count);
        Assert.IsTrue(messages[0].Flags.Contains(ImapMessageFlag.Seen));
        Assert.IsTrue(messages[0].IsSeen);
      });
    }

    [Test]
    public void TestStoreSequenceRangeSet()
    {
      TestMessage(3, ImapSequenceSet.CreateRangeSet(1L, 3L),
                  delegate(ImapPseudoServer server, ImapMessageInfoBase messageSet, ImapMessageInfo[] messages) {
        // STORE
        server.EnqueueTaggedResponse("* FETCH 1 (FLAGS (\\Draft))\r\n" +
                                     "* FETCH 2 (FLAGS ($label1))\r\n" +
                                     "* FETCH 3 (FLAGS ())\r\n" +
                                     "$tag OK done\r\n");

        messageSet.Store(ImapStoreDataItem.RemoveFlags(ImapMessageFlag.Seen));

        Assert.That(server.DequeueRequest(), Is.StringEnding("STORE 1:3 -FLAGS (\\Seen)\r\n"));

        Assert.IsNotNull(messages[0].Flags);
        Assert.AreEqual(1, messages[0].Flags.Count);
        Assert.IsTrue(messages[0].Flags.Contains(ImapMessageFlag.Draft));
        Assert.IsTrue(messages[0].IsDraft);

        Assert.IsNotNull(messages[1].Flags);
        Assert.AreEqual(1, messages[1].Flags.Count);
        Assert.IsTrue(messages[1].Flags.Contains("$label1"));

        Assert.IsNotNull(messages[2].Flags);
        Assert.AreEqual(0, messages[2].Flags.Count);
      });
    }

    [Test]
    public void TestStoreMailboxClosed()
    {
      TestMessage(10, ImapSequenceSet.CreateRangeSet(1L, 10L),
                  delegate(ImapPseudoServer server, ImapMessageInfoBase messageSet, ImapMessageInfo[] messages) {
        // CLOSE
        server.EnqueueTaggedResponse("$tag OK done\r\n");

        var mailbox = messageSet.Mailbox;

        messageSet.Mailbox.Close();

        Assert.IsFalse(mailbox.IsOpen);

        var ex = Assert.Throws<ImapUnavailableException>(() => messageSet.Store(ImapStoreDataItem.RemoveFlags(ImapMessageFlag.Seen)));

        Smdn.Net.TestUtils.SerializeBinary(ex);
      });
    }

    [Test, Ignore("not implemented; uid validity")]
    public void TestStoreUidValidityChanged()
    {
    }

    [Test]
    public void TestMarkAsDeleted()
    {
      TestMessage(3, ImapSequenceSet.CreateRangeSet(1L, 3L),
                  delegate(ImapPseudoServer server, ImapMessageInfoBase messageSet, ImapMessageInfo[] messages) {
        // STORE
        server.EnqueueTaggedResponse("* FETCH 1 (FLAGS (\\Deleted))\r\n" +
                                     "* FETCH 2 (FLAGS (\\Deleted \\Seen))\r\n" +
                                     "* FETCH 3 (FLAGS (\\Deleted $label1))\r\n" +
                                     "$tag OK done\r\n");

        messageSet.MarkAsDeleted();

        Assert.That(server.DequeueRequest(), Is.StringEnding("STORE 1:3 +FLAGS (\\Deleted)\r\n"));

        Assert.IsTrue(messages[0].IsAvailable);
        Assert.IsNotNull(messages[0].Flags);
        Assert.AreEqual(1, messages[0].Flags.Count);
        Assert.IsTrue(messages[0].Flags.Contains(ImapMessageFlag.Deleted));
        Assert.IsTrue(messages[0].IsMarkedAsDeleted);

        Assert.IsTrue(messages[1].IsAvailable);
        Assert.IsNotNull(messages[1].Flags);
        Assert.AreEqual(2, messages[1].Flags.Count);
        Assert.IsTrue(messages[1].Flags.Contains(ImapMessageFlag.Deleted));
        Assert.IsTrue(messages[1].Flags.Contains(ImapMessageFlag.Seen));
        Assert.IsTrue(messages[1].IsSeen);
        Assert.IsTrue(messages[1].IsMarkedAsDeleted);

        Assert.IsTrue(messages[2].IsAvailable);
        Assert.IsNotNull(messages[2].Flags);
        Assert.AreEqual(2, messages[2].Flags.Count);
        Assert.IsTrue(messages[2].Flags.Contains(ImapMessageFlag.Deleted));
        Assert.IsTrue(messages[2].Flags.Contains("$label1"));
        Assert.IsTrue(messages[2].IsMarkedAsDeleted);
      });
    }

    [Test]
    public void TestMarkAsSeen()
    {
      TestMessage(3, ImapSequenceSet.CreateRangeSet(1L, 3L),
                  delegate(ImapPseudoServer server, ImapMessageInfoBase messageSet, ImapMessageInfo[] messages) {
        // STORE
        server.EnqueueTaggedResponse("* FETCH 1 (FLAGS (\\Seen))\r\n" +
                                     "* FETCH 2 (FLAGS (\\Seen \\Draft))\r\n" +
                                     "* FETCH 3 (FLAGS (\\Seen $label1))\r\n" +
                                     "$tag OK done\r\n");

        messageSet.MarkAsSeen();

        Assert.That(server.DequeueRequest(), Is.StringEnding("STORE 1:3 +FLAGS (\\Seen)\r\n"));

        Assert.IsNotNull(messages[0].Flags);
        Assert.AreEqual(1, messages[0].Flags.Count);
        Assert.IsTrue(messages[0].Flags.Contains(ImapMessageFlag.Seen));
        Assert.IsTrue(messages[0].IsSeen);

        Assert.IsNotNull(messages[1].Flags);
        Assert.AreEqual(2, messages[1].Flags.Count);
        Assert.IsTrue(messages[1].Flags.Contains(ImapMessageFlag.Seen));
        Assert.IsTrue(messages[1].Flags.Contains(ImapMessageFlag.Draft));
        Assert.IsTrue(messages[1].IsSeen);
        Assert.IsTrue(messages[1].IsDraft);

        Assert.IsNotNull(messages[2].Flags);
        Assert.AreEqual(2, messages[2].Flags.Count);
        Assert.IsTrue(messages[2].Flags.Contains(ImapMessageFlag.Seen));
        Assert.IsTrue(messages[2].Flags.Contains("$label1"));
        Assert.IsTrue(messages[2].IsSeen);
      });
    }

    [Test]
    public void TestDeleteUidPlusCapable()
    {
      TestMessage(3,
                  new[] {ImapCapability.UidPlus},
                  ImapSequenceSet.CreateUidRangeSet(1L, 3L),
                  delegate(ImapPseudoServer server, ImapMessageInfoBase messageSet, ImapMessageInfo[] messages) {
        Delete(server, messageSet, messages, true);
      });
    }

    [Test]
    public void TestDeleteUidPlusIncapable()
    {
      TestMessage(3,
                  null,
                  ImapSequenceSet.CreateUidRangeSet(1L, 3L),
                  delegate(ImapPseudoServer server, ImapMessageInfoBase messageSet, ImapMessageInfo[] messages) {
        Delete(server, messageSet, messages, false);
      });
    }

    [Test]
    public void TestDeleteUidPlusDisabled()
    {
      TestMessage(3,
                  new[] {ImapCapability.UidPlus},
                  ImapSequenceSet.CreateUidRangeSet(1L, 3L),
                  delegate(ImapPseudoServer server, ImapMessageInfoBase messageSet, ImapMessageInfo[] messages) {
        messageSet.Client.DisabledCapabilities.Add(ImapCapability.UidPlus);

        Delete(server, messageSet, messages, false);
      });
    }

    private void Delete(ImapPseudoServer server, ImapMessageInfoBase messageSet, ImapMessageInfo[] messages, bool uidPlusAvailable)
    {
      // STORE
      server.EnqueueTaggedResponse("* FETCH 1 (FLAGS (\\Deleted))\r\n" +
                                   "* FETCH 2 (FLAGS (\\Deleted \\Seen))\r\n" +
                                   "* FETCH 3 (FLAGS (\\Deleted $label1))\r\n" +
                                   "$tag OK done\r\n");
      // EXPUNGE/UID EXPUNGE
      server.EnqueueTaggedResponse("* EXPUNGE 3\r\n" +
                                   "* EXPUNGE 2\r\n" +
                                   "* EXPUNGE 1\r\n" +
                                   "$tag OK done\r\n");

      var mailbox = messageSet.Mailbox;

      messageSet.Delete();

      Assert.That(server.DequeueRequest(), Is.StringEnding("STORE 1:3 +FLAGS (\\Deleted)\r\n"));
      if (uidPlusAvailable)
        Assert.That(server.DequeueRequest(), Is.StringEnding("UID EXPUNGE 1:3\r\n"));
      else
        Assert.That(server.DequeueRequest(), Is.StringEnding("EXPUNGE\r\n"));

      Assert.IsFalse(messageSet.IsAvailable);
      Assert.Throws<ImapUnavailableException>(() => Assert.IsNull(messageSet.Mailbox));
      Assert.AreEqual(0L, mailbox.ExistMessageCount);

      Assert.IsFalse(messages[0].IsAvailable);
      Assert.Throws<ImapUnavailableException>(() => Assert.IsNull(messages[0].Mailbox));
      Assert.AreEqual(ImapMessageInfo.ExpungedMessageSequenceNumber, messages[0].Sequence);
      Assert.IsTrue(messages[0].IsDeleted);

      Assert.IsNotNull(messages[0].Flags);
      Assert.AreEqual(1, messages[0].Flags.Count);
      Assert.IsTrue(messages[0].Flags.Contains(ImapMessageFlag.Deleted));
      Assert.IsTrue(messages[0].IsMarkedAsDeleted);

      Assert.IsFalse(messages[1].IsAvailable);
      Assert.Throws<ImapUnavailableException>(() => Assert.IsNull(messages[1].Mailbox));
      Assert.AreEqual(ImapMessageInfo.ExpungedMessageSequenceNumber, messages[1].Sequence);
      Assert.IsTrue(messages[1].IsDeleted);

      Assert.IsNotNull(messages[1].Flags);
      Assert.AreEqual(2, messages[1].Flags.Count);
      Assert.IsTrue(messages[1].Flags.Contains(ImapMessageFlag.Deleted));
      Assert.IsTrue(messages[1].Flags.Contains(ImapMessageFlag.Seen));
      Assert.IsTrue(messages[1].IsMarkedAsDeleted);
      Assert.IsTrue(messages[1].IsSeen);

      Assert.IsFalse(messages[0].IsAvailable);
      Assert.Throws<ImapUnavailableException>(() => Assert.IsNull(messages[2].Mailbox));
      Assert.AreEqual(ImapMessageInfo.ExpungedMessageSequenceNumber, messages[2].Sequence);
      Assert.IsTrue(messages[2].IsDeleted);

      Assert.IsNotNull(messages[2].Flags);
      Assert.AreEqual(2, messages[2].Flags.Count);
      Assert.IsTrue(messages[2].Flags.Contains(ImapMessageFlag.Deleted));
      Assert.IsTrue(messages[2].Flags.Contains("$label1"));
      Assert.IsTrue(messages[2].IsMarkedAsDeleted);
    }

    [Test]
    public void TestDeleteEmptySet()
    {
      TestMessage(0, CreateEmptySet(),
                  delegate(ImapPseudoServer server, ImapMessageInfoBase messageSet, ImapMessageInfo[] messages) {
        // do nothing
        messageSet.Delete();

        Assert.IsTrue(messageSet.IsAvailable);
        Assert.DoesNotThrow(() => Assert.IsNotNull(messageSet.Mailbox));
      });
    }

    [Test]
    public void TestCopyToDestinationMailbox()
    {
      TestMessage(3, ImapSequenceSet.CreateUidSet(1L, 3L, 5L),
                  delegate(ImapPseudoServer server, ImapMessageInfoBase messageSet, ImapMessageInfo[] messages) {
        CopyOrMoveToDestinationMailbox(server, messageSet, messages, false);
      });
    }

    [Test]
    public void TestMoveToDestinationMailbox()
    {
      TestMessage(3, ImapSequenceSet.CreateUidSet(1L, 3L, 5L),
                  delegate(ImapPseudoServer server, ImapMessageInfoBase messageSet, ImapMessageInfo[] messages) {
        CopyOrMoveToDestinationMailbox(server, messageSet, messages, true);
      });
    }

    [Test]
    public void TestMoveToMoveDisabledDestinationMailbox()
    {
      TestMessage(3, new[] {ImapCapability.Move}, ImapSequenceSet.CreateUidSet(1L, 3L, 5L),
                  delegate(ImapPseudoServer server, ImapMessageInfoBase messageSet, ImapMessageInfo[] messages) {
        messageSet.Client.DisabledCapabilities.Add(ImapCapability.Move);

        CopyOrMoveToDestinationMailbox(server, messageSet, messages, true);
      });
    }

    private void CopyOrMoveToDestinationMailbox(ImapPseudoServer server, ImapMessageInfoBase messageSet, ImapMessageInfo[] messages, bool move)
    {
      // LIST
      server.EnqueueTaggedResponse("* LIST () \"\" dest\r\n" +
                                   "$tag OK done\r\n");

      var destMailbox = messageSet.Client.GetMailbox("dest");

      server.DequeueRequest();

      // UID COPY
      server.EnqueueTaggedResponse("$tag OK done\r\n");
      if (move)
        // UID STORE
        server.EnqueueTaggedResponse("* FETCH 1 (FLAGS (\\Deleted))\r\n" +
                                     "* FETCH 2 (FLAGS (\\Deleted \\Seen))\r\n" +
                                     "* FETCH 3 (FLAGS (\\Deleted $label1))\r\n" +
                                     "$tag OK done\r\n");

      if (move)
        messageSet.MoveTo(destMailbox);
      else
        messageSet.CopyTo(destMailbox);

      Assert.That(server.DequeueRequest(), Is.StringEnding("UID COPY 1,3,5 dest\r\n"));

      if (move) {
        Assert.That(server.DequeueRequest(), Is.StringEnding("UID STORE 1,3,5 +FLAGS (\\Deleted)\r\n"));

        Assert.IsNotNull(messages[0].Flags);
        Assert.AreEqual(1, messages[0].Flags.Count);
        Assert.IsTrue(messages[0].Flags.Contains(ImapMessageFlag.Deleted));
        Assert.IsTrue(messages[0].IsMarkedAsDeleted);

        Assert.IsNotNull(messages[1].Flags);
        Assert.AreEqual(2, messages[1].Flags.Count);
        Assert.IsTrue(messages[1].Flags.Contains(ImapMessageFlag.Deleted));
        Assert.IsTrue(messages[1].Flags.Contains(ImapMessageFlag.Seen));
        Assert.IsTrue(messages[1].IsMarkedAsDeleted);
        Assert.IsTrue(messages[1].IsSeen);

        Assert.IsNotNull(messages[2].Flags);
        Assert.AreEqual(2, messages[2].Flags.Count);
        Assert.IsTrue(messages[2].Flags.Contains(ImapMessageFlag.Deleted));
        Assert.IsTrue(messages[2].Flags.Contains("$label1"));
        Assert.IsTrue(messages[2].IsMarkedAsDeleted);
      }
    }
    
    [Test]
    public void TestMoveToMoveCapableDestinationMailbox()
    {
      TestMessage(3,
                  new[] {ImapCapability.Move},
                  ImapSequenceSet.CreateUidRangeSet(1L, 3L),
                  delegate(ImapPseudoServer server, ImapMessageInfoBase messageSet, ImapMessageInfo[] messages) {
        // LIST
        server.EnqueueTaggedResponse("* LIST () \"\" dest\r\n" +
                                     "$tag OK done\r\n");

        var sourceMailbox = messageSet.Mailbox;
        var destMailbox = messageSet.Client.GetMailbox("dest");

        server.DequeueRequest();

        // UID MOVE
        server.EnqueueTaggedResponse("* EXPUNGE 3\r\n" +
                                     "* EXPUNGE 2\r\n" +
                                     "* EXPUNGE 1\r\n" +
                                     "$tag OK done\r\n");

        messageSet.MoveTo(destMailbox);

        Assert.That(server.DequeueRequest(), Is.StringEnding("UID MOVE 1:3 dest\r\n"));

        Assert.IsFalse(messageSet.IsAvailable);
        Assert.Throws<ImapUnavailableException>(() => Assert.IsNull(messageSet.Mailbox));
        Assert.AreEqual(0L, sourceMailbox.ExistMessageCount);

        Assert.IsFalse(messages[0].IsAvailable);
        Assert.AreEqual(ImapMessageInfo.ExpungedMessageSequenceNumber, messages[0].Sequence);
        Assert.IsTrue(messages[0].IsDeleted);

        Assert.IsFalse(messages[1].IsAvailable);
        Assert.AreEqual(ImapMessageInfo.ExpungedMessageSequenceNumber, messages[1].Sequence);
        Assert.IsTrue(messages[1].IsDeleted);

        Assert.IsFalse(messages[2].IsAvailable);
        Assert.AreEqual(ImapMessageInfo.ExpungedMessageSequenceNumber, messages[2].Sequence);
        Assert.IsTrue(messages[2].IsDeleted);
      });
    }

    [Test]
    public void TestMoveToDestinationMailboxCopyFailed()
    {
      MoveToDestinationMailboxCopyFailed(false);
    }

    [Test]
    public void TestMoveToMoveCapableDestinationMailboxCopyFailed()
    {
      MoveToDestinationMailboxCopyFailed(true);
    }

    private void MoveToDestinationMailboxCopyFailed(bool moveCapable)
    {
      var capabilities = moveCapable ? new[] { ImapCapability.Move } : EmptyArray<ImapCapability>.Instance;

      TestMessage(3, capabilities, ImapSequenceSet.CreateUidSet(1L, 3L, 5L),
                  delegate(ImapPseudoServer server, ImapMessageInfoBase messageSet, ImapMessageInfo[] messages) {
        // LIST
        server.EnqueueTaggedResponse("* LIST () \"\" dest\r\n" +
                                     "$tag OK done\r\n");

        var destMailbox = messageSet.Client.GetMailbox("dest");

        server.DequeueRequest();

        // UID COPY/UID MOVE
        server.EnqueueTaggedResponse("$tag NO failed\r\n");

        TestUtils.ExpectExceptionThrown<ImapErrorResponseException>(delegate {
          // STORE won't be issued
          messageSet.MoveTo(destMailbox);
        });

        Assert.IsTrue(messageSet.IsAvailable);

        if (moveCapable)
          Assert.That(server.DequeueRequest(), Is.StringEnding("UID MOVE 1,3,5 dest\r\n"));
        else
          Assert.That(server.DequeueRequest(), Is.StringEnding("UID COPY 1,3,5 dest\r\n"));
      });
    }

    [Test]
    public void TestCopyToDestinationMailboxItself()
    {
      CopyOrMoveToDestinationMailboxItself(false, EmptyArray<ImapCapability>.Instance);
    }

    [Test]
    public void TestMoveToDestinationMailboxItself()
    {
      CopyOrMoveToDestinationMailboxItself(true, EmptyArray<ImapCapability>.Instance);
    }

    [Test]
    public void TestMoveToMoveCapableDestinationMailboxItself()
    {
      CopyOrMoveToDestinationMailboxItself(true, new[] {ImapCapability.Move});
    }

    private void CopyOrMoveToDestinationMailboxItself(bool move, ImapCapability[] capabilities)
    {
      TestMessage(3, capabilities, ImapSequenceSet.CreateUidSet(1L, 3L, 5L),
                  delegate(ImapPseudoServer server, ImapMessageInfoBase messageSet, ImapMessageInfo[] messages) {
        Assert.AreEqual(3, messageSet.Mailbox.ExistMessageCount);
        Assert.AreEqual(0, messageSet.Mailbox.RecentMessageCount);

        if (!move) {
          // UID COPY
          server.EnqueueTaggedResponse("* 3 RECENT\r\n" +
                                       "* 6 EXISTS\r\n" +
                                       "$tag OK done\r\n");
        }

        if (move)
          messageSet.MoveTo(messageSet.Mailbox);
        else
          messageSet.CopyTo(messageSet.Mailbox);

        Assert.IsTrue(messageSet.IsAvailable);

        if (move) {
          Assert.AreEqual(3, messageSet.Mailbox.ExistMessageCount);
          Assert.AreEqual(0, messageSet.Mailbox.RecentMessageCount);
        }
        else {
          Assert.That(server.DequeueRequest(), Is.StringEnding("UID COPY 1,3,5 " + messageSet.Mailbox.FullName + "\r\n"));

          Assert.AreEqual(6, messageSet.Mailbox.ExistMessageCount);
          Assert.AreEqual(3, messageSet.Mailbox.RecentMessageCount);
        }
      });
    }

    [Test]
    public void TestCopyToDestinationMailboxOfDifferentSession()
    {
      CopyOrMoveToDestinationMailboxOfDifferentSession(false, EmptyArray<ImapCapability>.Instance);
    }

    [Test]
    public void TestMoveToDestinationMailboxOfDifferentSession()
    {
      CopyOrMoveToDestinationMailboxOfDifferentSession(true, EmptyArray<ImapCapability>.Instance);
    }

    [Test]
    public void TestMoveToMoveCapableDestinationMailboxOfDifferentSession()
    {
      CopyOrMoveToDestinationMailboxOfDifferentSession(true, new[] {ImapCapability.Move});
    }

    private void CopyOrMoveToDestinationMailboxOfDifferentSession(bool move, ImapCapability[] capabilities)
    {
      TestMessage(3, capabilities, ImapSequenceSet.CreateUidSet(1L, 3L, 5L),
                  delegate(ImapPseudoServer serverSrc, ImapMessageInfoBase srcMessageSet, ImapMessageInfo[] srcMessages) {
        TestUtils.TestAuthenticated(delegate(ImapPseudoServer serverDest, ImapClient clientDest) {
          // LIST
          serverDest.EnqueueTaggedResponse("* LIST () \"\" dest\r\n" +
                                           "$tag OK done\r\n");

          var destMailbox = clientDest.GetMailbox("dest");

          serverDest.DequeueRequest();

          Assert.AreNotSame(srcMessageSet.Client, destMailbox.Client);

          // FETCH
          serverSrc.EnqueueTaggedResponse("$tag NO FETCH failed\r\n");

          TestUtils.ExpectExceptionThrown<ImapErrorResponseException>(delegate {
            if (move)
              srcMessageSet.MoveTo(destMailbox);
            else
              srcMessageSet.CopyTo(destMailbox);
          });

          Assert.That(serverSrc.DequeueRequest(), Is.StringEnding("UID FETCH 1,3,5 (UID)\r\n"));
        });
      });
    }

    [Test, Ignore("not implemented; uid validity")]
    public void TestCopyToDestinationMailboxUidValidityChanged()
    {
    }

    [Test, Ignore("not implemented; uid validity")]
    public void TestMoveToDestinationMailboxUidValidityChanged()
    {
    }

    [Test, Ignore("not implemented; uid validity")]
    public void TestMoveToMoveCapableDestinationMailboxUidValidityChanged()
    {
    }

    [Test]
    public void TestCopyToDestinationMailboxArgumentNull()
    {
      CopyOrMoveToDestinationMailboxArgumentNull(false, EmptyArray<ImapCapability>.Instance);
    }

    [Test]
    public void TestMoveToDestinationMailboxArgumentNull()
    {
      CopyOrMoveToDestinationMailboxArgumentNull(true, EmptyArray<ImapCapability>.Instance);
    }

    [Test]
    public void TestMoveToMoveCapableDestinationMailboxArgumentNull()
    {
      CopyOrMoveToDestinationMailboxArgumentNull(true, new[] {ImapCapability.Move});
    }

    private void CopyOrMoveToDestinationMailboxArgumentNull(bool move, ImapCapability[] capabilities)
    {
      TestMessage(3, capabilities, ImapSequenceSet.CreateUidSet(1L, 3L, 5L),
                  delegate(ImapPseudoServer server, ImapMessageInfoBase messageSet, ImapMessageInfo[] messages) {
        if (move)
          Assert.Throws<ArgumentNullException>(() => messageSet.MoveTo((ImapMailboxInfo)null));
        else
          Assert.Throws<ArgumentNullException>(() => messageSet.CopyTo((ImapMailboxInfo)null));
      });
    }

    [Test]
    public void TestCopyToDestinationMailboxNonExistent()
    {
      CopyOrMoveToDestinationMailboxNonExistent(false, EmptyArray<ImapCapability>.Instance);
    }

    [Test]
    public void TestMoveToDestinationMailboxNonExistent()
    {
      CopyOrMoveToDestinationMailboxNonExistent(true, EmptyArray<ImapCapability>.Instance);
    }

    [Test]
    public void TestMoveToMoveCapableDestinationMailboxNonExistent()
    {
      CopyOrMoveToDestinationMailboxNonExistent(true, new[] {ImapCapability.Move});
    }

    private void CopyOrMoveToDestinationMailboxNonExistent(bool move, ImapCapability[] capabilities)
    {
      TestMessage(10,
                  capabilities.Prepend(ImapCapability.ListExtended),
                  ImapSequenceSet.CreateUidRangeSet(1L, 10L),
                  delegate(ImapPseudoServer server, ImapMessageInfoBase messageSet, ImapMessageInfo[] messages) {
        // LIST
        server.EnqueueTaggedResponse("* LIST (\\Subscribed \\NonExistent) \"\" dest\r\n" +
                                     "$tag OK done\r\n");

        var destMailbox = messageSet.Client.GetMailbox("dest", ImapMailboxListOptions.SubscribedOnly);

        server.DequeueRequest();

        Assert.IsFalse(destMailbox.Exists);

        if (move)
          Assert.Throws<ImapProtocolViolationException>(() => messageSet.MoveTo(destMailbox));
        else
          Assert.Throws<ImapProtocolViolationException>(() => messageSet.CopyTo(destMailbox));
      });
    }

    [Test]
    public void TestCopyToDestinationMailboxUnselectable()
    {
      CopyOrMoveToDestinationMailboxUnselectable(false, EmptyArray<ImapCapability>.Instance);
    }

    [Test]
    public void TestMoveToDestinationMailboxUnselectable()
    {
      CopyOrMoveToDestinationMailboxUnselectable(true, EmptyArray<ImapCapability>.Instance);
    }

    [Test]
    public void TestMoveToMoveCapableDestinationMailboxUnselectable()
    {
      CopyOrMoveToDestinationMailboxUnselectable(true, new[] {ImapCapability.Move});
    }

    private void CopyOrMoveToDestinationMailboxUnselectable(bool move, ImapCapability[] capabilities)
    {
      TestMessage(10,
                  capabilities.Prepend(ImapCapability.ListExtended),
                  ImapSequenceSet.CreateUidRangeSet(1L, 10L),
                  delegate(ImapPseudoServer server, ImapMessageInfoBase messageSet, ImapMessageInfo[] messages) {
        // LIST
        server.EnqueueTaggedResponse("* LIST (\\Noselect) \"\" dest\r\n" +
                                     "$tag OK done\r\n");

        var destMailbox = messageSet.Client.GetMailbox("dest", ImapMailboxListOptions.SubscribedOnly);

        server.DequeueRequest();

        Assert.IsTrue(destMailbox.IsUnselectable);

        if (move)
          Assert.Throws<ImapProtocolViolationException>(() => messageSet.MoveTo(destMailbox));
        else
          Assert.Throws<ImapProtocolViolationException>(() => messageSet.CopyTo(destMailbox));
      });
    }

    [Test]
    public void TestCopyToDestinationMailboxEmptySet()
    {
      CopyOrMoveToDestinationMailboxEmptySet(false, EmptyArray<ImapCapability>.Instance);
    }

    [Test]
    public void TestMoveToDestinationMailboxEmptySet()
    {
      CopyOrMoveToDestinationMailboxEmptySet(true, EmptyArray<ImapCapability>.Instance);
    }

    [Test]
    public void TestMoveToMoveCapableDestinationMailboxEmptySet()
    {
      CopyOrMoveToDestinationMailboxEmptySet(true, new[] {ImapCapability.Move});
    }

    private void CopyOrMoveToDestinationMailboxEmptySet(bool move, ImapCapability[] capabilities)
    {
      TestMessage(0, capabilities, CreateEmptySet(),
                  delegate(ImapPseudoServer server, ImapMessageInfoBase messageSet, ImapMessageInfo[] messages) {
        // LIST
        server.EnqueueTaggedResponse("* LIST () \"\" dest\r\n" +
                                     "$tag OK done\r\n");

        var destMailbox = messageSet.Client.GetMailbox("dest");

        server.DequeueRequest();

        // do nothing
        if (move)
          messageSet.MoveTo(destMailbox);
        else
          messageSet.CopyTo(destMailbox);

        Assert.IsTrue(messageSet.IsAvailable);
      });
    }

    [Test]
    public void TestCopyToDestinationMailboxMailboxClosed()
    {
      CopyOrMoveToDestinationMailboxMailboxClosed(false, EmptyArray<ImapCapability>.Instance);
    }

    [Test]
    public void TestMoveToDestinationMailboxMailboxClosed()
    {
      CopyOrMoveToDestinationMailboxMailboxClosed(true, EmptyArray<ImapCapability>.Instance);
    }

    [Test]
    public void TestMoveToMoveCapableDestinationMailboxMailboxClosed()
    {
      CopyOrMoveToDestinationMailboxMailboxClosed(true, new[] {ImapCapability.Move});
    }

    private void CopyOrMoveToDestinationMailboxMailboxClosed(bool move, ImapCapability[] capabilities)
    {
      TestMessage(3, capabilities, ImapSequenceSet.CreateUidSet(1L, 3L, 5L),
                  delegate(ImapPseudoServer server, ImapMessageInfoBase messageSet, ImapMessageInfo[] messages) {
        // LIST
        server.EnqueueTaggedResponse("* LIST () \"\" dest\r\n" +
                                     "$tag OK done\r\n");

        var destMailbox = messageSet.Client.GetMailbox("dest");

        server.DequeueRequest();

        // CLOSE
        server.EnqueueTaggedResponse("$tag OK done\r\n");

        var mailbox = messageSet.Mailbox;

        mailbox.Close();

        Assert.IsFalse(mailbox.IsOpen);
        Assert.IsFalse(messageSet.IsAvailable);
        Assert.Throws<ImapUnavailableException>(() => Assert.IsNull(messageSet.Mailbox));

        server.DequeueRequest();

        var ex = Assert.Throws<ImapUnavailableException>(delegate {
          if (move)
            messageSet.MoveTo(destMailbox);
          else
            messageSet.CopyTo(destMailbox);
        });

        Smdn.Net.TestUtils.SerializeBinary(ex);
      });
    }

    [Test]
    public void TestCopyToDestinationName()
    {
      TestMessage(3, ImapSequenceSet.CreateUidRangeSet(1L, 3L),
                  delegate(ImapPseudoServer server, ImapMessageInfoBase messageSet, ImapMessageInfo[] messages) {
        CopyOrMoveToDestinationName(server, messageSet, messages, false, false);
      });
    }

    [Test]
    public void TestMoveToDestinationName()
    {
      TestMessage(3, ImapSequenceSet.CreateUidRangeSet(1L, 3L),
                  delegate(ImapPseudoServer server, ImapMessageInfoBase messageSet, ImapMessageInfo[] messages) {
        CopyOrMoveToDestinationName(server, messageSet, messages, true, false);
      });
    }

    [Test]
    public void TestMoveToMoveDisabledDestinationName()
    {
      TestMessage(3, new[] {ImapCapability.Move}, ImapSequenceSet.CreateUidRangeSet(1L, 3L),
                  delegate(ImapPseudoServer server, ImapMessageInfoBase messageSet, ImapMessageInfo[] messages) {
        messageSet.Client.DisabledCapabilities.Add(ImapCapability.Move);

        CopyOrMoveToDestinationName(server, messageSet, messages, true, false);
      });
    }

    [Test]
    public void TestCopyToDestinationNameTryCreate()
    {
      TestMessage(3, ImapSequenceSet.CreateUidRangeSet(1L, 3L),
                  delegate(ImapPseudoServer server, ImapMessageInfoBase messageSet, ImapMessageInfo[] messages) {
        CopyOrMoveToDestinationName(server, messageSet, messages, false, true);
      });
    }

    [Test]
    public void TestMoveToDestinationNameTryCreate()
    {
      TestMessage(3, ImapSequenceSet.CreateUidRangeSet(1L, 3L),
                  delegate(ImapPseudoServer server, ImapMessageInfoBase messageSet, ImapMessageInfo[] messages) {
        CopyOrMoveToDestinationName(server, messageSet, messages, true, true);
      });
    }

    private void CopyOrMoveToDestinationName(ImapPseudoServer server, ImapMessageInfoBase messageSet, ImapMessageInfo[] messages, bool move, bool tryCreate)
    {
      // UID COPY
      server.EnqueueTaggedResponse("$tag OK done\r\n");
      if (move)
        // UID STORE
        server.EnqueueTaggedResponse("* FETCH 1 (FLAGS (\\Deleted))\r\n" +
                                     "* FETCH 2 (FLAGS (\\Deleted \\Seen))\r\n" +
                                     "* FETCH 3 (FLAGS (\\Deleted $label1))\r\n" +
                                     "$tag OK done\r\n");

      ImapMailboxInfo createdMailbox = null;

      if (tryCreate) {
        createdMailbox = move
          ? messageSet.MoveTo("dest", true)
          : messageSet.CopyTo("dest", true);
      }
      else {
        if (move)
          messageSet.MoveTo("dest");
        else
          messageSet.CopyTo("dest");
      }

      if (tryCreate)
        Assert.IsNull(createdMailbox);

      Assert.That(server.DequeueRequest(), Is.StringEnding("UID COPY 1:3 dest\r\n"));

      if (move) {
        Assert.That(server.DequeueRequest(), Is.StringEnding("UID STORE 1:3 +FLAGS (\\Deleted)\r\n"));

        Assert.IsFalse(messageSet.IsAvailable);

        Assert.IsNotNull(messages[0].Flags);
        Assert.AreEqual(1, messages[0].Flags.Count);
        Assert.IsTrue(messages[0].Flags.Contains(ImapMessageFlag.Deleted));
        Assert.IsTrue(messages[0].IsMarkedAsDeleted);

        Assert.IsNotNull(messages[1].Flags);
        Assert.AreEqual(2, messages[1].Flags.Count);
        Assert.IsTrue(messages[1].Flags.Contains(ImapMessageFlag.Deleted));
        Assert.IsTrue(messages[1].Flags.Contains(ImapMessageFlag.Seen));
        Assert.IsTrue(messages[1].IsMarkedAsDeleted);
        Assert.IsTrue(messages[1].IsSeen);

        Assert.IsNotNull(messages[2].Flags);
        Assert.AreEqual(2, messages[2].Flags.Count);
        Assert.IsTrue(messages[2].Flags.Contains(ImapMessageFlag.Deleted));
        Assert.IsTrue(messages[2].Flags.Contains("$label1"));
        Assert.IsTrue(messages[2].IsMarkedAsDeleted);
      }
    }

    [Test]
    public void TestMoveToMoveCapableDestinationName()
    {
      MoveToMoveCapableDestinationName(false);
    }

    [Test]
    public void TestMoveToMoveCapableDestinationNameNoResponseTryCreate()
    {
      MoveToMoveCapableDestinationName(true);
    }

    private void MoveToMoveCapableDestinationName(bool tryCreate)
    {
      TestMessage(3,
                  new[] {ImapCapability.Move},
                  ImapSequenceSet.CreateUidRangeSet(1L, 3L),
                  delegate(ImapPseudoServer server, ImapMessageInfoBase messageSet, ImapMessageInfo[] messages) {
        if (tryCreate) {
          // UID MOVE
          server.EnqueueTaggedResponse("$tag NO [TRYCREATE] mailbox not exist\r\n");
          // CREATE
          server.EnqueueTaggedResponse("$tag OK done\r\n");
          // UID MOVE
          server.EnqueueTaggedResponse("* EXPUNGE 3\r\n" +
                                       "* EXPUNGE 2\r\n" +
                                       "* EXPUNGE 1\r\n" +
                                       "$tag OK done\r\n");
          // LIST
          server.EnqueueTaggedResponse("* LIST () \"\" dest\r\n" +
                                       "$tag OK done\r\n");
        }
        else {
          // UID MOVE
          server.EnqueueTaggedResponse("* EXPUNGE 3\r\n" +
                                       "* EXPUNGE 2\r\n" +
                                       "* EXPUNGE 1\r\n" +
                                       "$tag OK done\r\n");
        }

        var sourceMailbox = messageSet.Mailbox;
        var createdMailbox = messageSet.MoveTo("dest", tryCreate);

        if (tryCreate) {
          Assert.IsNotNull(createdMailbox);
          Assert.AreEqual("dest", createdMailbox.Name);
          Assert.AreEqual("dest", createdMailbox.FullName);
        }
        else {
          Assert.IsNull(createdMailbox);
        }

        Assert.That(server.DequeueRequest(), Is.StringEnding("UID MOVE 1:3 dest\r\n"));

        Assert.IsFalse(messageSet.IsAvailable);
        Assert.Throws<ImapUnavailableException>(() => Assert.IsNull(messageSet.Mailbox));
        Assert.AreEqual(0L, sourceMailbox.ExistMessageCount);

        Assert.IsFalse(messages[0].IsAvailable);
        Assert.AreEqual(ImapMessageInfo.ExpungedMessageSequenceNumber, messages[0].Sequence);
        Assert.IsTrue(messages[0].IsDeleted);

        Assert.IsFalse(messages[1].IsAvailable);
        Assert.AreEqual(ImapMessageInfo.ExpungedMessageSequenceNumber, messages[1].Sequence);
        Assert.IsTrue(messages[1].IsDeleted);

        Assert.IsFalse(messages[2].IsAvailable);
        Assert.AreEqual(ImapMessageInfo.ExpungedMessageSequenceNumber, messages[2].Sequence);
        Assert.IsTrue(messages[2].IsDeleted);
      });
    }

    [Test]
    public void TestCopyToDestinationNameSameMailbox()
    {
      CopyOrMoveToDestinationNameSameMailbox(false, EmptyArray<ImapCapability>.Instance);
    }

    [Test]
    public void TestMoveToDestinationNameSameMailbox()
    {
      CopyOrMoveToDestinationNameSameMailbox(true, EmptyArray<ImapCapability>.Instance);
    }

    [Test]
    public void TestMoveToMoveCapableDestinationNameSameMailbox()
    {
      CopyOrMoveToDestinationNameSameMailbox(true, new[] {ImapCapability.Move});
    }

    private void CopyOrMoveToDestinationNameSameMailbox(bool move, ImapCapability[] capabilities)
    {
      TestMessage(3, capabilities, ImapSequenceSet.CreateUidRangeSet(1L, 3L),
                  delegate(ImapPseudoServer server, ImapMessageInfoBase messageSet, ImapMessageInfo[] messages) {
        Assert.AreEqual(3, messageSet.Mailbox.ExistMessageCount);
        Assert.AreEqual(0, messageSet.Mailbox.RecentMessageCount);

        if (!move) {
          // UID COPY
          server.EnqueueTaggedResponse("* 3 RECENT\r\n" +
                                       "* 6 EXISTS\r\n" +
                                       "$tag OK done\r\n");
        }

        if (move)
          messageSet.MoveTo(messageSet.Mailbox.FullName);
        else
          messageSet.CopyTo(messageSet.Mailbox.FullName);

        Assert.IsTrue(messageSet.IsAvailable);

        if (move) {
          Assert.AreEqual(3, messageSet.Mailbox.ExistMessageCount);
          Assert.AreEqual(0, messageSet.Mailbox.RecentMessageCount);
        }
        else {
          Assert.That(server.DequeueRequest(), Is.StringEnding("UID COPY 1:3 " + messageSet.Mailbox.FullName + "\r\n"));

          Assert.AreEqual(6, messageSet.Mailbox.ExistMessageCount);
          Assert.AreEqual(3, messageSet.Mailbox.RecentMessageCount);
        }
      });
    }

    [Test]
    public void TestCopyToDestinationNameArgumentEmpty()
    {
      CopyOrMoveToDestinationNameArgumentEmpty(false, EmptyArray<ImapCapability>.Instance);
    }

    [Test]
    public void TestMoveToDestinationNameArgumentEmpty()
    {
      CopyOrMoveToDestinationNameArgumentEmpty(true, EmptyArray<ImapCapability>.Instance);
    }

    [Test]
    public void TestMoveToMoveCapableDestinationNameArgumentEmpty()
    {
      CopyOrMoveToDestinationNameArgumentEmpty(true, new[] {ImapCapability.Move});
    }

    private void CopyOrMoveToDestinationNameArgumentEmpty(bool move, ImapCapability[] capabilities)
    {
      TestMessage(3, capabilities, ImapSequenceSet.CreateUidSet(1L, 3L, 5L),
                  delegate(ImapPseudoServer server, ImapMessageInfoBase messageSet, ImapMessageInfo[] messages) {
        if (move)
          Assert.Throws<ArgumentException>(() => messageSet.MoveTo(string.Empty));
        else
          Assert.Throws<ArgumentException>(() => messageSet.CopyTo(string.Empty));
      });
    }

    [Test]
    public void TestCopyToDestinationNameArgumentNull()
    {
      CopyOrMoveToDestinationNameArgumentNull(false, EmptyArray<ImapCapability>.Instance);
    }

    [Test]
    public void TestMoveToDestinationNameArgumentNull()
    {
      CopyOrMoveToDestinationNameArgumentNull(true, EmptyArray<ImapCapability>.Instance);
    }

    [Test]
    public void TestMoveToMoveCapableDestinationNameArgumentNull()
    {
      CopyOrMoveToDestinationNameArgumentNull(true, new[] {ImapCapability.Move});
    }

    private void CopyOrMoveToDestinationNameArgumentNull(bool move, ImapCapability[] capabilities)
    {
      TestMessage(3, capabilities, ImapSequenceSet.CreateUidSet(1L, 3L, 5L),
                  delegate(ImapPseudoServer server, ImapMessageInfoBase messageSet, ImapMessageInfo[] messages) {
        if (move)
          Assert.Throws<ArgumentNullException>(() => messageSet.MoveTo((string)null));
        else
          Assert.Throws<ArgumentNullException>(() => messageSet.CopyTo((string)null));
      });
    }

    [Test]
    public void TestCopyToDestinationNameEmptySet()
    {
      CopyOrMoveToDestinationNameEmptySet(false, EmptyArray<ImapCapability>.Instance);
    }

    [Test]
    public void TestMoveToDestinationNameEmptySet()
    {
      CopyOrMoveToDestinationNameEmptySet(true, EmptyArray<ImapCapability>.Instance);
    }

    [Test]
    public void TestMoveToMoveCapableDestinationNameEmptySet()
    {
      CopyOrMoveToDestinationNameEmptySet(true, new[] {ImapCapability.Move});
    }

    private void CopyOrMoveToDestinationNameEmptySet(bool move, ImapCapability[] capabilities)
    {
      TestMessage(0, capabilities, CreateEmptySet(),
                  delegate(ImapPseudoServer server, ImapMessageInfoBase messageSet, ImapMessageInfo[] messages) {
        // do nothing
        if (move)
          messageSet.MoveTo("dest");
        else
          messageSet.CopyTo("dest");

        Assert.IsTrue(messageSet.IsAvailable);
      });
    }

    [Test, Ignore("not implemented; uid validity")]
    public void TestCopyToDestinationNameUidValidityChanged()
    {
    }

    [Test, Ignore("not implemented; uid validity")]
    public void TestMoveToDestinationNameUidValidityChanged()
    {
    }

    [Test, Ignore("not implemented; uid validity")]
    public void TestMoveToMoveCapableDestinationNameUidValidityChanged()
    {
    }

    [Test]
    public void TestCopyToDestinationNameMailboxClosed()
    {
      CopyOrMoveToDestinationNameMailboxClosed(false, EmptyArray<ImapCapability>.Instance);
    }

    [Test]
    public void TestMoveToDestinationNameMailboxClosed()
    {
      CopyOrMoveToDestinationNameMailboxClosed(true, EmptyArray<ImapCapability>.Instance);
    }

    [Test]
    public void TestMoveToMoveCapableDestinationNameMailboxClosed()
    {
      CopyOrMoveToDestinationNameMailboxClosed(true, new[] {ImapCapability.Move});
    }

    private void CopyOrMoveToDestinationNameMailboxClosed(bool move, ImapCapability[] capabilities)
    {
      TestMessage(10, capabilities, ImapSequenceSet.CreateUidRangeSet(1L, 10L),
                  delegate(ImapPseudoServer server, ImapMessageInfoBase messageSet, ImapMessageInfo[] messages) {
        // CLOSE
        server.EnqueueTaggedResponse("$tag OK done\r\n");

        var mailbox = messageSet.Mailbox;

        mailbox.Close();

        Assert.IsFalse(mailbox.IsOpen);
        Assert.IsFalse(messageSet.IsAvailable);
        Assert.Throws<ImapUnavailableException>(() => Assert.IsNull(messageSet.Mailbox));

        server.DequeueRequest();

        var ex = Assert.Throws<ImapUnavailableException>(delegate {
          if (move)
            messageSet.MoveTo("dest");
          else
            messageSet.CopyTo("dest");
        });

        Smdn.Net.TestUtils.SerializeBinary(ex);
      });
    }

    [Test]
    public void TestCopyToDestinationNameNoResponse()
    {
      CopyOrMoveToDestinationNameNoResponse(false, EmptyArray<ImapCapability>.Instance);
    }

    [Test]
    public void TestMoveToDestinationNameNoResponse()
    {
      CopyOrMoveToDestinationNameNoResponse(true, EmptyArray<ImapCapability>.Instance);
    }

    [Test]
    public void TestMoveToMoveCapableDestinationNameNoResponse()
    {
      CopyOrMoveToDestinationNameNoResponse(true, new[] {ImapCapability.Move});
    }

    private void CopyOrMoveToDestinationNameNoResponse(bool move, ImapCapability[] capabilities)
    {
      TestMessage(10, capabilities, ImapSequenceSet.CreateUidRangeSet(1L, 10L),
                  delegate(ImapPseudoServer server, ImapMessageInfoBase messageSet, ImapMessageInfo[] messages) {
        // UID COPY
        server.EnqueueTaggedResponse("$tag NO error occurred\r\n");

        if (move)
          Assert.Throws<ImapErrorResponseException>(() => messageSet.MoveTo("dest", false));
        else
          Assert.Throws<ImapErrorResponseException>(() => messageSet.CopyTo("dest", false));
      });
    }

    [Test]
    public void TestCopyToDestinationNameNoResponseDontTryCreate()
    {
      CopyOrMoveToDestinationNameNoResponseDontTryCreate(false, EmptyArray<ImapCapability>.Instance);
    }

    [Test]
    public void TestMoveToDestinationNameNoResponseDontTryCreate()
    {
      CopyOrMoveToDestinationNameNoResponseDontTryCreate(true, EmptyArray<ImapCapability>.Instance);
    }

    [Test]
    public void TestMoveToMoveCapableDestinationNameNoResponseDontTryCreate()
    {
      CopyOrMoveToDestinationNameNoResponseDontTryCreate(true, new[] {ImapCapability.Move});
    }

    private void CopyOrMoveToDestinationNameNoResponseDontTryCreate(bool move, ImapCapability[] capabilities)
    {
      TestMessage(10, capabilities, ImapSequenceSet.CreateUidRangeSet(1L, 10L),
                  delegate(ImapPseudoServer server, ImapMessageInfoBase messageSet, ImapMessageInfo[] messages) {
        // UID COPY
        server.EnqueueTaggedResponse("$tag NO [TRYCREATE] mailbox not exist\r\n");

        try {
          if (move)
            messageSet.MoveTo("dest", false);
          else
            messageSet.CopyTo("dest", false);

          Assert.Fail("ImapMailboxNotFoundException not thrown");
        }
        catch (ImapMailboxNotFoundException ex) {
          Assert.IsNotNull(ex.Mailbox);
          Assert.AreEqual("dest", ex.Mailbox);

          Smdn.Net.TestUtils.SerializeBinary(ex, delegate(ImapMailboxNotFoundException deserialized) {
            Assert.IsNotNull(deserialized.Mailbox);
            Assert.AreEqual(ex.Mailbox, deserialized.Mailbox);
          });
        }

        Assert.IsTrue(messageSet.IsAvailable);
      });
    }

    [Test]
    public void TestCopyToDestinationNameNoResponseTryCreate()
    {
      CopyOrMoveToDestinationNameNoResponseTryCreate(false);
    }

    [Test]
    public void TestMoveToDestinationNameNoResponseTryCreate()
    {
      CopyOrMoveToDestinationNameNoResponseTryCreate(true);
    }

    private void CopyOrMoveToDestinationNameNoResponseTryCreate(bool move)
    {
      TestMessage(3, ImapSequenceSet.CreateUidRangeSet(1L, 3L),
                  delegate(ImapPseudoServer server, ImapMessageInfoBase messageSet, ImapMessageInfo[] messages) {
        // UID COPY
        server.EnqueueTaggedResponse("$tag NO [TRYCREATE] mailbox not exist\r\n");
        // CREATE
        server.EnqueueTaggedResponse("$tag OK done\r\n");
        // UID COPY
        server.EnqueueTaggedResponse("$tag OK done\r\n");
        // LIST
        server.EnqueueTaggedResponse("* LIST () \"\" dest\r\n" +
                                     "$tag OK done\r\n");
        if (move)
          // UID STORE
          server.EnqueueTaggedResponse("* FETCH 1 (FLAGS (\\Deleted))\r\n" +
                                       "* FETCH 2 (FLAGS (\\Deleted \\Seen))\r\n" +
                                       "* FETCH 3 (FLAGS (\\Deleted $label1))\r\n" +
                                       "$tag OK done\r\n");

        var createdMailbox = move
          ? messageSet.MoveTo("dest", true)
          : messageSet.CopyTo("dest", true);

        Assert.IsNotNull(createdMailbox);
        Assert.AreEqual("dest", createdMailbox.Name);
        Assert.AreEqual("dest", createdMailbox.FullName);

        Assert.That(server.DequeueRequest(), Is.StringEnding("UID COPY 1:3 dest\r\n"));
        Assert.That(server.DequeueRequest(), Is.StringEnding("CREATE dest\r\n"));
        Assert.That(server.DequeueRequest(), Is.StringEnding("UID COPY 1:3 dest\r\n"));
        Assert.That(server.DequeueRequest(), Is.StringEnding("LIST \"\" dest\r\n"));

        if (move) {
          Assert.That(server.DequeueRequest(), Is.StringEnding("UID STORE 1:3 +FLAGS (\\Deleted)\r\n"));

          Assert.IsFalse(messageSet.IsAvailable);

          Assert.IsNotNull(messages[0].Flags);
          Assert.AreEqual(1, messages[0].Flags.Count);
          Assert.IsTrue(messages[0].Flags.Contains(ImapMessageFlag.Deleted));
          Assert.IsTrue(messages[0].IsMarkedAsDeleted);

          Assert.IsNotNull(messages[1].Flags);
          Assert.AreEqual(2, messages[1].Flags.Count);
          Assert.IsTrue(messages[1].Flags.Contains(ImapMessageFlag.Deleted));
          Assert.IsTrue(messages[1].Flags.Contains(ImapMessageFlag.Seen));
          Assert.IsTrue(messages[1].IsMarkedAsDeleted);
          Assert.IsTrue(messages[1].IsSeen);

          Assert.IsNotNull(messages[2].Flags);
          Assert.AreEqual(2, messages[2].Flags.Count);
          Assert.IsTrue(messages[2].Flags.Contains(ImapMessageFlag.Deleted));
          Assert.IsTrue(messages[2].Flags.Contains("$label1"));
          Assert.IsTrue(messages[2].IsMarkedAsDeleted);
        }
      });
    }

    [Test]
    public void TestCopyToDestinationNameNoResponseTryCreateNameContainsWildcard()
    {
      CopyOrMoveToDestinationNameNoResponseTryCreateNameContainsWildcard(true);
    }

    [Test]
    public void TestMoveToDestinationNameNoResponseTryCreateNameContainsWildcard()
    {
      CopyOrMoveToDestinationNameNoResponseTryCreateNameContainsWildcard(false);
    }

    private void CopyOrMoveToDestinationNameNoResponseTryCreateNameContainsWildcard(bool move)
    {
      TestMessage(3, ImapSequenceSet.CreateUidRangeSet(1L, 3L),
                  delegate(ImapPseudoServer server, ImapMessageInfoBase messageSet, ImapMessageInfo[] messages) {
        // UID COPY
        server.EnqueueTaggedResponse("$tag NO [TRYCREATE] mailbox not exist\r\n");
        // CREATE
        server.EnqueueTaggedResponse("$tag OK done\r\n");
        // UID COPY
        server.EnqueueTaggedResponse("$tag OK done\r\n");
        // LIST
        server.EnqueueTaggedResponse("* LIST (\\Marked) \".\" Box*\r\n" +
                                     "* LIST () \".\" Box\r\n" +
                                     "* LIST () \".\" Box.Child\r\n" +
                                     "* LIST () \".\" Boxes\r\n" +
                                     "$tag OK done\r\n");
        if (move)
          // UID STORE
          server.EnqueueTaggedResponse("* FETCH 1 (FLAGS (\\Deleted))\r\n" +
                                       "* FETCH 2 (FLAGS (\\Deleted \\Seen))\r\n" +
                                       "* FETCH 3 (FLAGS (\\Deleted $label1))\r\n" +
                                       "$tag OK done\r\n");

        var createdMailbox = move
          ? messageSet.MoveTo("Box*", true)
          : messageSet.CopyTo("Box*", true);

        Assert.IsNotNull(createdMailbox);
        Assert.AreEqual("Box*", createdMailbox.Name);
        Assert.AreEqual("Box*", createdMailbox.FullName);
        Assert.AreEqual(1, createdMailbox.Flags.Count);
        CollectionAssert.Contains(createdMailbox.Flags, ImapMailboxFlag.Marked);

        Assert.That(server.DequeueRequest(), Is.StringEnding("UID COPY 1:3 \"Box*\"\r\n"));
        Assert.That(server.DequeueRequest(), Is.StringEnding("CREATE \"Box*\"\r\n"));
        Assert.That(server.DequeueRequest(), Is.StringEnding("UID COPY 1:3 \"Box*\"\r\n"));
        Assert.That(server.DequeueRequest(), Is.StringEnding("LIST \"\" Box*\r\n"));

        if (move) {
          Assert.That(server.DequeueRequest(), Is.StringEnding("UID STORE 1:3 +FLAGS (\\Deleted)\r\n"));

          Assert.IsFalse(messageSet.IsAvailable);

          Assert.IsNotNull(messages[0].Flags);
          Assert.AreEqual(1, messages[0].Flags.Count);
          Assert.IsTrue(messages[0].Flags.Contains(ImapMessageFlag.Deleted));
          Assert.IsTrue(messages[0].IsMarkedAsDeleted);

          Assert.IsNotNull(messages[1].Flags);
          Assert.AreEqual(2, messages[1].Flags.Count);
          Assert.IsTrue(messages[1].Flags.Contains(ImapMessageFlag.Deleted));
          Assert.IsTrue(messages[1].Flags.Contains(ImapMessageFlag.Seen));
          Assert.IsTrue(messages[1].IsMarkedAsDeleted);
          Assert.IsTrue(messages[1].IsSeen);

          Assert.IsNotNull(messages[2].Flags);
          Assert.AreEqual(2, messages[2].Flags.Count);
          Assert.IsTrue(messages[2].Flags.Contains(ImapMessageFlag.Deleted));
          Assert.IsTrue(messages[2].Flags.Contains("$label1"));
          Assert.IsTrue(messages[2].IsMarkedAsDeleted);
        }
      });
    }
  }
}
