using System;
using System.Collections.Generic;
using System.Linq;
using System.Threading;
using NUnit.Framework;

using Smdn.Net.Imap4.Protocol;

namespace Smdn.Net.Imap4.Client {
  [TestFixture]
  public class ImapOpenedMailboxInfoOperationsWaitingTests {
    public static void TestOpenedMailboxDisconnectable(ImapCapability[] capabilities, string mailboxName, string selectResponse, Action<ImapPseudoServer, ImapOpenedMailboxInfo> action)
    {
      TestUtils.TestAuthenticated(capabilities, delegate(ImapPseudoServer server, ImapClient client) {
        // LIST
        server.EnqueueTaggedResponse(string.Format("* LIST () \"\" {0}\r\n", mailboxName) +
                                     "$tag OK done\r\n");
        // SELECT
        server.EnqueueTaggedResponse(selectResponse);

        var opened = client.OpenMailbox(mailboxName);

        server.DequeueRequest(); // LIST
        server.DequeueRequest(); // SELECT

        action(server, opened);
      });
    }

    [Test]
    public void TestIdleIdleCapable()
    {
      TestUtils.TestOpenedMailbox(new[] {ImapCapability.Idle},
                                  "INBOX",
                                  "$tag OK [READ-WRITE] done\r\n",
                                  delegate(ImapPseudoServer server, ImapOpenedMailboxInfo mailbox) {
        using (var waitHandle = new ManualResetEvent(false)) {
          var timeout = TimeSpan.FromSeconds(0.5);

          ThreadPool.QueueUserWorkItem(delegate(object state) {
            waitHandle.Set();

            // IDLE
            server.EnqueueResponse("+ idling\r\n");

            Thread.Sleep(timeout + TimeSpan.FromSeconds(0.25));

            Assert.IsTrue(mailbox.Client.IsBusy, "IsBusy while idling");

            // DONE
            server.EnqueueTaggedResponse("$tag OK done\r\n");
          });

          Smdn.Net.TestUtils.Elapses(timeout, delegate {
            waitHandle.WaitOne();
            waitHandle.Reset();

            Assert.IsFalse(mailbox.Client.IsBusy, "IsBusy before idle");

            mailbox.Idle(timeout);
          });

          Assert.IsFalse(mailbox.Client.IsBusy, "IsBusy after idle");
        }
      });
    }

    [Test]
    public void TestIdleIdleCapableTimeoutZero()
    {
      TestUtils.TestOpenedMailbox(new[] {ImapCapability.Idle},
                                  "INBOX",
                                  "$tag OK [READ-WRITE] done\r\n",
                                  delegate(ImapPseudoServer server, ImapOpenedMailboxInfo mailbox) {
        using (var waitHandle = new ManualResetEvent(false)) {
          var delay = TimeSpan.FromSeconds(0.1);

          ThreadPool.QueueUserWorkItem(delegate(object state) {
            waitHandle.Set();

            // IDLE
            server.EnqueueResponse("+ idling\r\n");

            Thread.Sleep(delay);

            // DONE
            server.EnqueueTaggedResponse("$tag OK done\r\n");
          });

          Smdn.Net.TestUtils.Elapses(delay, delegate {
            waitHandle.WaitOne();
            waitHandle.Reset();

            mailbox.Idle(0);
          });
        }
      });
    }

    [Test]
    public void TestIdleIdleCapableDisconnectFromServer()
    {
      TestOpenedMailboxDisconnectable(new[] {ImapCapability.Idle},
                                      "INBOX",
                                      "$tag OK [READ-WRITE] done\r\n",
                                      delegate(ImapPseudoServer server, ImapOpenedMailboxInfo mailbox) {
        using (var waitHandle = new ManualResetEvent(false)) {
          var timeout = TimeSpan.FromSeconds(1.5);

          ThreadPool.QueueUserWorkItem(delegate(object state) {
            waitHandle.Set();

            // IDLE
            server.EnqueueResponse("+ idling\r\n");

            Thread.Sleep(TimeSpan.FromSeconds(0.25));

            server.EnqueueResponse("* BYE Server shutting down.\r\n");
          });

          Smdn.Net.TestUtils.NotElapse(timeout, delegate {
            waitHandle.WaitOne();
            waitHandle.Reset();

            Assert.Throws<ImapConnectionException>(() => mailbox.Idle(timeout));
          });

          server.Stop();
        }
      });
    }

    [Test]
    public void TestIdleIdleIncapable()
    {
      TestUtils.TestOpenedMailbox("INBOX",
                                  "$tag OK [READ-WRITE] done\r\n",
                                  IdleUnavailable);
    }

    [Test]
    public void TestIdleIdleDisabled()
    {
      TestUtils.TestOpenedMailbox(new[] {ImapCapability.Idle},
                                  "INBOX",
                                  "$tag OK [READ-WRITE] done\r\n",
                                  delegate(ImapPseudoServer server, ImapOpenedMailboxInfo mailbox) {
        mailbox.Client.DisabledCapabilities.Add(ImapCapability.Idle);

        IdleUnavailable(server, mailbox);
      });
    }

    private void IdleUnavailable(ImapPseudoServer server, ImapOpenedMailboxInfo mailbox)
    {
      // NOOP (initial)
      server.EnqueueTaggedResponse("$tag OK done\r\n");
      // NOOP (first polling)
      server.EnqueueTaggedResponse("$tag OK done\r\n");
      // NOOP (stopped)
      server.EnqueueTaggedResponse("$tag OK done\r\n");

      var timeout = TimeSpan.FromSeconds(0.5);

      Smdn.Net.TestUtils.Elapses(timeout, delegate {
        Assert.IsFalse(mailbox.Client.IsBusy, "IsBusy before idle");

        mailbox.Idle(timeout, TimeSpan.FromSeconds(0.35));
      });

      Assert.IsFalse(mailbox.Client.IsBusy, "IsBusy after idle");
    }

    [Test]
    public void TestIdleIdleIncapableTimeoutZero()
    {
      TestUtils.TestOpenedMailbox("INBOX",
                                  "$tag OK [READ-WRITE] done\r\n",
                                  delegate(ImapPseudoServer server, ImapOpenedMailboxInfo mailbox) {
        // NOOP (initial)
        server.EnqueueTaggedResponse("$tag OK done\r\n");
        // NOOP (stopped)
        server.EnqueueTaggedResponse("$tag OK done\r\n");

        Smdn.Net.TestUtils.Elapses(TimeSpan.Zero, delegate {
          mailbox.Idle(0, 1000);
        });
      });
    }

    [Test]
    public void TestIdleIdleIncapableDisconnectFromServer()
    {
      TestOpenedMailboxDisconnectable(new ImapCapability[0],
                                      "INBOX",
                                      "$tag OK [READ-WRITE] done\r\n",
                                      delegate(ImapPseudoServer server, ImapOpenedMailboxInfo mailbox) {
        // NOOP (initial)
        server.EnqueueTaggedResponse("$tag OK done\r\n");
        // NOOP (first polling)
        server.EnqueueResponse("* BYE Server shutting down.\r\n");

        var timeout = TimeSpan.FromSeconds(2.5);
        var pollingInterval = TimeSpan.FromSeconds(0.25);

        Smdn.Net.TestUtils.Elapses(pollingInterval, delegate {
          Assert.Throws<ImapConnectionException>(() => mailbox.Idle(timeout, pollingInterval));
        });

        server.Stop();
      });
    }

    [Test]
    public void TestIdleStopIdleWhileAwaiting()
    {
      TestUtils.TestOpenedMailbox(new[] {ImapCapability.Idle},
                                  "INBOX",
                                  "$tag OK [READ-WRITE] done\r\n",
                                  delegate(ImapPseudoServer server, ImapOpenedMailboxInfo mailbox) {
        using (var waitHandle = new ManualResetEvent(false)) {
          var timeout = TimeSpan.FromSeconds(2.5);

          ThreadPool.QueueUserWorkItem(delegate(object state) {
            waitHandle.Set();

            // IDLE
            server.EnqueueResponse("+ idling\r\n");

            Thread.Sleep(TimeSpan.FromSeconds(0.1));

            // DONE
            server.EnqueueTaggedResponse("$tag OK done\r\n");

            try {
              mailbox.StopIdle();
            }
            catch {
              // ignore exceptions
            }
          });

          Smdn.Net.TestUtils.NotElapse(timeout, delegate {
            waitHandle.WaitOne();
            waitHandle.Reset();

            Assert.Throws<ImapProtocolViolationException>(() => mailbox.Idle(timeout));
          });

          Assert.IsFalse(mailbox.Client.IsBusy);
        }
      });
    }

    [Test]
    public void TestIdleTimeoutInfinite()
    {
      TestUtils.TestOpenedMailbox("INBOX",
                                  "$tag OK [READ-WRITE] done\r\n",
                                  delegate(ImapPseudoServer server, ImapOpenedMailboxInfo mailbox) {
        Assert.Throws<ArgumentOutOfRangeException>(() => mailbox.Idle(Timeout.Infinite));
      });
    }

    [Test]
    public void TestIdleTimeoutArgumentOutOfRange()
    {
      TestUtils.TestOpenedMailbox("INBOX",
                                  "$tag OK [READ-WRITE] done\r\n",
                                  delegate(ImapPseudoServer server, ImapOpenedMailboxInfo mailbox) {
        Assert.Throws<ArgumentOutOfRangeException>(() => mailbox.Idle(-2));
      });
    }

    [Test]
    public void TestStartIdleIdleCapablePollingIntervalInifinite()
    {
      StartIdleIdleCapablePollingIntervalIgnored(Timeout.Infinite);
    }

    [Test]
    public void TestStartIdleIdleCapablePollingIntervalArgumentOutOfRange1()
    {
      StartIdleIdleCapablePollingIntervalIgnored(-2);
    }

    [Test]
    public void TestStartIdleIdleCapablePollingIntervalArgumentOutOfRange2()
    {
      StartIdleIdleCapablePollingIntervalIgnored((int)TimeSpan.FromMinutes(30.0).TotalMilliseconds);
    }

    private void StartIdleIdleCapablePollingIntervalIgnored(int millisecondsPollingInterval)
    {
      TestUtils.TestOpenedMailbox(new[] {ImapCapability.Idle},
                                  "INBOX",
                                  "$tag OK [READ-WRITE] done\r\n",
                                  delegate(ImapPseudoServer server, ImapOpenedMailboxInfo mailbox) {
        using (var waitHandle = new ManualResetEvent(false)) {
          var timeout = TimeSpan.FromSeconds(0.25);

          ThreadPool.QueueUserWorkItem(delegate(object state) {
            waitHandle.Set();

            // IDLE
            server.EnqueueResponse("+ idling\r\n");

            waitHandle.WaitOne();

            // DONE
            server.EnqueueTaggedResponse("$tag OK done\r\n");
          });

          waitHandle.WaitOne();
          waitHandle.Reset();

          Smdn.Net.TestUtils.Elapses(timeout, delegate {
            Assert.IsFalse(mailbox.Client.IsBusy, "IsBusy before idle");

            mailbox.StartIdle(millisecondsPollingInterval);

            Thread.Sleep(25); // XXX: wait for idle started

            Assert.IsTrue(mailbox.Client.IsBusy, "IsBusy while idling");

            Thread.Sleep(timeout);

            waitHandle.Set();

            mailbox.StopIdle();

            Assert.IsFalse(mailbox.Client.IsBusy, "IsBusy after idle");
          });
        }
      });
    }

    [Test]
    public void TestStartIdleIdleIncapablePollingIntervalInfinite()
    {
      TestUtils.TestOpenedMailbox("INBOX",
                                  "$tag OK [READ-WRITE] done\r\n",
                                  delegate(ImapPseudoServer server, ImapOpenedMailboxInfo mailbox) {
        Assert.Throws<ArgumentOutOfRangeException>(() => mailbox.StartIdle(Timeout.Infinite));
      });
    }

    [Test]
    public void TestStartIdleIdleIncapablePollingIntervalArgumentOutOfRange1()
    {
      TestUtils.TestOpenedMailbox("INBOX",
                                  "$tag OK [READ-WRITE] done\r\n",
                                  delegate(ImapPseudoServer server, ImapOpenedMailboxInfo mailbox) {
        Assert.Throws<ArgumentOutOfRangeException>(() => mailbox.StartIdle(-2));
      });
    }

    [Test]
    public void TestStartIdleIdleIncapablePollingIntervalArgumentOutOfRange2()
    {
      TestUtils.TestOpenedMailbox("INBOX",
                                  "$tag OK [READ-WRITE] done\r\n",
                                  delegate(ImapPseudoServer server, ImapOpenedMailboxInfo mailbox) {
        Assert.Throws<ArgumentOutOfRangeException>(() => mailbox.StartIdle(TimeSpan.FromMinutes(30.0)));
      });
    }

    [Test]
    public void TestStartIdleAnotherWatchOperationProceeding1()
    {
      TestUtils.TestOpenedMailbox(new[] {ImapCapability.Idle},
                                  "INBOX",
                                  "$tag OK [READ-WRITE] done\r\n",
                                  delegate(ImapPseudoServer server, ImapOpenedMailboxInfo mailbox) {
        Assert.IsFalse(mailbox.Client.IsBusy, "IsBusy before idle");

        // first
        try {
          server.EnqueueResponse("+ idling\r\n");

          mailbox.StartIdle();
        }
        catch (Exception ex) {
          Assert.Fail("unexpected exception thrown: {0}", ex);
        }

        Thread.Sleep(25); // XXX: wait for idle started

        Assert.IsTrue(mailbox.Client.IsBusy, "IsBusy while idling");

        // second
        Assert.Throws<ImapProtocolViolationException>(() => mailbox.StartIdle());

        server.EnqueueTaggedResponse("$tag OK done\r\n");

        mailbox.StopIdle();

        Assert.IsFalse(mailbox.Client.IsBusy, "IsBusy after idle");
      });
    }

    [Test]
    public void TestStartIdleAnotherWatchOperationProceeding2()
    {
      TestUtils.TestOpenedMailbox(new[] {ImapCapability.Idle},
                                  "INBOX",
                                  "$tag OK [READ-WRITE] done\r\n",
                                  delegate(ImapPseudoServer server, ImapOpenedMailboxInfo mailbox) {
        Assert.IsFalse(mailbox.Client.IsBusy, "IsBusy before idle");

        // first
        server.EnqueueResponse("+ idling\r\n");

        mailbox.StartWaitForMessageArrival();

        Thread.Sleep(25); // XXX: wait for idle started

        Assert.IsTrue(mailbox.Client.IsBusy, "IsBusy while idling");

        // second
        Assert.Throws<ImapProtocolViolationException>(() => mailbox.StartIdle());

        server.EnqueueTaggedResponse("$tag OK done\r\n");

        mailbox.StopWaitForMessageArrival();

        Assert.IsFalse(mailbox.Client.IsBusy, "IsBusy after idle");
      });
    }

    [Test]
    public void TestStopIdleWatchOperationNotStarted()
    {
      TestUtils.TestOpenedMailbox(new[] {ImapCapability.Idle},
                                  "INBOX",
                                  "$tag OK [READ-WRITE] done\r\n",
                                  delegate(ImapPseudoServer server, ImapOpenedMailboxInfo mailbox) {
        Assert.Throws<ImapProtocolViolationException>(() => mailbox.StopIdle());
      });
    }

    [Test]
    public void TestStopIdleAnotherWatchOperationProceeding()
    {
      TestUtils.TestOpenedMailbox(new[] {ImapCapability.Idle},
                                  "INBOX",
                                  "$tag OK [READ-WRITE] done\r\n",
                                  delegate(ImapPseudoServer server, ImapOpenedMailboxInfo mailbox) {
        server.EnqueueResponse("+ idling\r\n");

        mailbox.StartWaitForMessageArrival();

        Thread.Sleep(25); // XXX: wait for idle started

        Assert.IsTrue(mailbox.Client.IsBusy);

        Assert.Throws<ImapProtocolViolationException>(() => mailbox.StopIdle());

        server.EnqueueTaggedResponse("$tag OK done\r\n");

        mailbox.StopWaitForMessageArrival();

        Assert.IsFalse(mailbox.Client.IsBusy);
      });
    }

    [Test]
    public void TestWaitForMessageArrivalIdleCapable()
    {
      var selectResp =
        "* 0 EXISTS\r\n" +
        "* OK [UIDVALIDITY 1]\r\n" +
        "$tag OK [READ-WRITE] done\r\n";

      TestUtils.TestOpenedMailbox(new[] {ImapCapability.Idle},
                                  "INBOX",
                                  selectResp,
                                  delegate(ImapPseudoServer server, ImapOpenedMailboxInfo mailbox) {
        using (var waitHandle = new ManualResetEvent(false)) {
          var timeout = TimeSpan.FromSeconds(0.5);

          ThreadPool.QueueUserWorkItem(delegate(object state) {
            waitHandle.Set();

            // IDLE
            server.EnqueueResponse("+ idling\r\n");

            Thread.Sleep(timeout + TimeSpan.FromSeconds(0.25));

            // DONE
            server.EnqueueTaggedResponse("$tag OK done\r\n");
          });

          ImapMessageInfoList arrivalMessages = null;

          Smdn.Net.TestUtils.Elapses(timeout, delegate {
            waitHandle.WaitOne();
            waitHandle.Reset();

            arrivalMessages = mailbox.WaitForMessageArrival(timeout);
          });

          Assert.IsNotNull(arrivalMessages);
          CollectionAssert.IsEmpty(arrivalMessages);
        }
      });
    }

    [Test]
    public void TestWaitForMessageArrivalIdleCapableTimeoutZero()
    {
      var selectResp =
        "* 0 EXISTS\r\n" +
        "* OK [UIDVALIDITY 1]\r\n" +
        "$tag OK [READ-WRITE] done\r\n";

      TestUtils.TestOpenedMailbox(new[] {ImapCapability.Idle},
                                  "INBOX",
                                  selectResp,
                                  delegate(ImapPseudoServer server, ImapOpenedMailboxInfo mailbox) {
        using (var waitHandle = new ManualResetEvent(false)) {
          var delay = TimeSpan.FromSeconds(0.1);

          ThreadPool.QueueUserWorkItem(delegate(object state) {
            waitHandle.Set();

            // IDLE
            server.EnqueueResponse("+ idling\r\n");

            Thread.Sleep(delay);

            // DONE
            server.EnqueueTaggedResponse("$tag OK done\r\n");
          });

          ImapMessageInfoList arrivalMessages = null;

          Smdn.Net.TestUtils.Elapses(delay, delegate {
            waitHandle.WaitOne();
            waitHandle.Reset();

            arrivalMessages = mailbox.WaitForMessageArrival(0);
          });

          Assert.IsNotNull(arrivalMessages);
          CollectionAssert.IsEmpty(arrivalMessages);
        }
      });
    }

    [Test]
    public void TestWaitForMessageArrivalIdleCapableTimeoutInfinite()
    {
      var selectResp =
        "* 0 EXISTS\r\n" +
        "* OK [UIDVALIDITY 1]\r\n" +
        "$tag OK [READ-WRITE] done\r\n";

      TestUtils.TestOpenedMailbox(new[] {ImapCapability.Idle},
                                  "INBOX",
                                  selectResp,
                                  delegate(ImapPseudoServer server, ImapOpenedMailboxInfo mailbox) {
        using (var waitHandle = new ManualResetEvent(false)) {
          var delay = TimeSpan.FromSeconds(0.1);

          ThreadPool.QueueUserWorkItem(delegate(object state) {
            waitHandle.Set();

            // IDLE
            server.EnqueueResponse("+ idling\r\n");

            Thread.Sleep(delay);

            server.EnqueueResponse("* RECENT 1\r\n");
            server.EnqueueResponse("* EXISTS 1\r\n");

            Thread.Sleep(TimeSpan.FromSeconds(0.1));

            // DONE
            server.EnqueueTaggedResponse("$tag OK done\r\n");
          });

          ImapMessageInfoList arrivalMessages = null;

          Smdn.Net.TestUtils.Elapses(delay, delegate {
            waitHandle.WaitOne();
            waitHandle.Reset();

            arrivalMessages = mailbox.WaitForMessageArrival(Timeout.Infinite);
          });

          Assert.IsNotNull(arrivalMessages);

          // FETCH
          server.EnqueueTaggedResponse("* FETCH 1 (UID 1)\r\n" +
                                       "$tag OK done\r\n");

          var arrivalMessageArray = arrivalMessages.ToArray();

          Assert.AreEqual(1, arrivalMessageArray.Length);
          Assert.AreEqual(1L, arrivalMessageArray[0].Uid, "arrival message uid");
          Assert.AreEqual(1L, arrivalMessageArray[0].Sequence, "arrival message seq");
        }
      });
    }

    [Test]
    public void TestWaitForMessageArrivalIdleIncapable()
    {
      var selectResp =
        "* 0 EXISTS\r\n" +
        "* OK [UIDVALIDITY 1]\r\n" +
        "$tag OK [READ-WRITE] done\r\n";

      TestUtils.TestOpenedMailbox("INBOX",
                                  selectResp,
                                  delegate(ImapPseudoServer server, ImapOpenedMailboxInfo mailbox) {
        server.WaitForRequest = true;
        // NOOP (initial)
        server.EnqueueTaggedResponse("$tag OK done\r\n");
        // NOOP (first polling)
        server.EnqueueTaggedResponse("$tag OK done\r\n");
        // NOOP (stopped)
        server.EnqueueTaggedResponse("$tag OK done\r\n");

        var timeout = TimeSpan.FromSeconds(0.5);
        ImapMessageInfoList arrivalMessages = null;

        Smdn.Net.TestUtils.Elapses(timeout, delegate {
          arrivalMessages = mailbox.WaitForMessageArrival(timeout,
                                                          TimeSpan.FromSeconds(0.35));
        });

        Assert.IsNotNull(arrivalMessages);
        CollectionAssert.IsEmpty(arrivalMessages);
      });
    }

    [Test]
    public void TestWaitForMessageArrivalIdleIncapableTimeoutZero()
    {
      var selectResp =
        "* 0 EXISTS\r\n" +
        "* OK [UIDVALIDITY 1]\r\n" +
        "$tag OK [READ-WRITE] done\r\n";

      TestUtils.TestOpenedMailbox("INBOX",
                                  selectResp,
                                  delegate(ImapPseudoServer server, ImapOpenedMailboxInfo mailbox) {
        // NOOP (initial)
        server.EnqueueTaggedResponse("$tag OK done\r\n");
        // NOOP (first polling)
        server.EnqueueTaggedResponse("$tag OK done\r\n");

        var pollingInterval = TimeSpan.FromSeconds(0.25);
        ImapMessageInfoList arrivalMessages = null;

        Smdn.Net.TestUtils.NotElapse(pollingInterval, delegate {
          arrivalMessages = mailbox.WaitForMessageArrival(0,
                                                          (int)pollingInterval.TotalMilliseconds);
        });

        Assert.IsNotNull(arrivalMessages);

        var arrivalMessageArray = arrivalMessages.ToArray();

        Assert.AreEqual(0, arrivalMessageArray.Length);
      });
    }

    [Test]
    public void TestWaitForMessageArrivalIdleIncapableTimeoutInfinite()
    {
      var selectResp =
        "* 0 EXISTS\r\n" +
        "* OK [UIDVALIDITY 1]\r\n" +
        "$tag OK [READ-WRITE] done\r\n";

      TestUtils.TestOpenedMailbox("INBOX",
                                  selectResp,
                                  delegate(ImapPseudoServer server, ImapOpenedMailboxInfo mailbox) {
        // NOOP (initial)
        server.EnqueueTaggedResponse("$tag OK done\r\n");
        // NOOP (first polling)
        server.EnqueueTaggedResponse("$tag OK done\r\n");
        // NOOP (second polling)
        server.EnqueueTaggedResponse("* RECENT 1\r\n" +
                                     "* EXISTS 1\r\n" +
                                     "$tag OK done\r\n");

        var pollingInterval = TimeSpan.FromSeconds(0.25);
        ImapMessageInfoList arrivalMessages = null;

        Smdn.Net.TestUtils.Elapses(TimeSpan.FromTicks(pollingInterval.Ticks * 2), delegate {
          arrivalMessages = mailbox.WaitForMessageArrival(Timeout.Infinite,
                                                          (int)pollingInterval.TotalMilliseconds);
        });

        Assert.IsNotNull(arrivalMessages);

        // FETCH
        server.EnqueueTaggedResponse("* FETCH 1 (UID 1)\r\n" +
                                     "$tag OK done\r\n");

        var arrivalMessageArray = arrivalMessages.ToArray();

        Assert.AreEqual(1, arrivalMessageArray.Length);
        Assert.AreEqual(1L, arrivalMessageArray[0].Uid, "arrival message uid");
        Assert.AreEqual(1L, arrivalMessageArray[0].Sequence, "arrival message seq");
      });
    }

    [Test]
    public void TestWaitForMessageArrivalTimeoutArgumentOutOfRange()
    {
      TestUtils.TestOpenedMailbox(new[] {ImapCapability.Idle},
                                  "INBOX",
                                  "$tag OK [READ-WRITE] done\r\n",
                                  delegate(ImapPseudoServer server, ImapOpenedMailboxInfo mailbox) {
        Assert.Throws<ArgumentOutOfRangeException>(() => mailbox.WaitForMessageArrival(-2));
      });
    }

    [Test]
    public void TestWaitForMessageArrivalStopWaitForMessageArrivalWhileAwaiting()
    {
      var selectResp =
        "* 0 EXISTS\r\n" +
        "* OK [UIDVALIDITY 1]\r\n" +
        "$tag OK [READ-WRITE] done\r\n";

      TestUtils.TestOpenedMailbox(new[] {ImapCapability.Idle},
                                  "INBOX",
                                  selectResp,
                                  delegate(ImapPseudoServer server, ImapOpenedMailboxInfo mailbox) {
        using (var waitHandle = new ManualResetEvent(false)) {
          var timeout = TimeSpan.FromSeconds(2.5);

          ThreadPool.QueueUserWorkItem(delegate(object state) {
            waitHandle.Set();

            // IDLE
            server.EnqueueResponse("+ idling\r\n");

            Thread.Sleep(TimeSpan.FromSeconds(0.25));

            // DONE
            server.EnqueueTaggedResponse("$tag OK done\r\n");

            mailbox.StopWaitForMessageArrival();
          });

          Smdn.Net.TestUtils.NotElapse(timeout, delegate {
            waitHandle.WaitOne();
            waitHandle.Reset();

            Assert.Throws<ImapProtocolViolationException>(() => mailbox.WaitForMessageArrival(timeout));
          });
        }
      });
    }

    [Test]
    public void TestStartWaitForMessageArrivalIdleCapablePollingIntervalInifinite()
    {
      StartWaitForMessageArrivalIdleCapablePollingIntervalIgnored(-1);
    }

    [Test]
    public void TestStartWaitForMessageArrivalIdleCapablePollingIntervalArgumentOutOfRange1()
    {
      StartWaitForMessageArrivalIdleCapablePollingIntervalIgnored(-2);
    }

    [Test]
    public void TestStartWaitForMessageArrivalIdleCapablePollingIntervalArgumentOutOfRange2()
    {
      StartWaitForMessageArrivalIdleCapablePollingIntervalIgnored((int)TimeSpan.FromMinutes(30.0).TotalMilliseconds);
    }

    private void StartWaitForMessageArrivalIdleCapablePollingIntervalIgnored(int millisecondsPollingInterval)
    {
      TestUtils.TestOpenedMailbox(new[] {ImapCapability.Idle},
                                  "INBOX",
                                  "$tag OK [READ-WRITE] done\r\n",
                                  delegate(ImapPseudoServer server, ImapOpenedMailboxInfo mailbox) {
        using (var waitHandle = new ManualResetEvent(false)) {
          var timeout = TimeSpan.FromSeconds(0.25);

          ThreadPool.QueueUserWorkItem(delegate(object state) {
            waitHandle.Set();

            // IDLE
            server.EnqueueResponse("+ idling\r\n");

            waitHandle.WaitOne();

            // DONE
            server.EnqueueTaggedResponse("$tag OK done\r\n");
          });

          waitHandle.WaitOne();
          waitHandle.Reset();

          Smdn.Net.TestUtils.Elapses(timeout, delegate {
            mailbox.StartWaitForMessageArrival(millisecondsPollingInterval);

            Thread.Sleep(timeout);

            waitHandle.Set();

            mailbox.StopWaitForMessageArrival();
          });
        }
      });
    }

    [Test]
    public void TestStartWaitForMessageArrivalIdleIncapablePollingIntervalInfinite()
    {
      TestUtils.TestOpenedMailbox("INBOX",
                                  "$tag OK [READ-WRITE] done\r\n",
                                  delegate(ImapPseudoServer server, ImapOpenedMailboxInfo mailbox) {
        Assert.Throws<ArgumentOutOfRangeException>(() => mailbox.StartWaitForMessageArrival(Timeout.Infinite));
      });
    }

    [Test]
    public void TestStartWaitForMessageArrivalIdleIncapablePollingIntervalArgumentOutOfRange1()
    {
      TestUtils.TestOpenedMailbox("INBOX",
                                  "$tag OK [READ-WRITE] done\r\n",
                                  delegate(ImapPseudoServer server, ImapOpenedMailboxInfo mailbox) {
        Assert.Throws<ArgumentOutOfRangeException>(() => mailbox.StartWaitForMessageArrival(-2));
      });
    }

    [Test]
    public void TestStartWaitForMessageArrivalIdleIncapablePollingIntervalArgumentOutOfRange2()
    {
      TestUtils.TestOpenedMailbox("INBOX",
                                  "$tag OK [READ-WRITE] done\r\n",
                                  delegate(ImapPseudoServer server, ImapOpenedMailboxInfo mailbox) {
        Assert.Throws<ArgumentOutOfRangeException>(() => mailbox.StartWaitForMessageArrival(TimeSpan.FromMinutes(30.0)));
      });
    }

    [Test]
    public void TestStartWaitForMessageArrivalAnotherWatchOperationProceeding1()
    {
      TestUtils.TestOpenedMailbox(new[] {ImapCapability.Idle},
                                  "INBOX",
                                  "$tag OK [READ-WRITE] done\r\n",
                                  delegate(ImapPseudoServer server, ImapOpenedMailboxInfo mailbox) {

        Assert.IsFalse(mailbox.Client.IsBusy, "IsBusy before idle");

        // first
        server.EnqueueResponse("+ idling\r\n");

        mailbox.StartWaitForMessageArrival();

        Thread.Sleep(25); // XXX: wait for idle started

        Assert.IsTrue(mailbox.Client.IsBusy, "IsBusy while idling");

        // second
        Assert.Throws<ImapProtocolViolationException>(() => mailbox.StartWaitForMessageArrival());

        server.EnqueueTaggedResponse("$tag OK done\r\n");

        mailbox.StopWaitForMessageArrival();

        Assert.IsFalse(mailbox.Client.IsBusy, "IsBusy after idle");
      });
    }

    [Test]
    public void TestStartWaitForMessageArrivalAnotherWatchOperationProceeding2()
    {
      TestUtils.TestOpenedMailbox(new[] {ImapCapability.Idle},
                                  "INBOX",
                                  "$tag OK [READ-WRITE] done\r\n",
                                  delegate(ImapPseudoServer server, ImapOpenedMailboxInfo mailbox) {
        Assert.IsFalse(mailbox.Client.IsBusy, "IsBusy before idle");

        // first
        server.EnqueueResponse("+ idling\r\n");

        mailbox.StartIdle();

        Thread.Sleep(25); // XXX: wait for idle started

        Assert.IsTrue(mailbox.Client.IsBusy, "IsBusy while idling");

        // second
        Assert.Throws<ImapProtocolViolationException>(() => mailbox.StartWaitForMessageArrival());

        server.EnqueueTaggedResponse("$tag OK done\r\n");

        mailbox.StopIdle();

        Assert.IsFalse(mailbox.Client.IsBusy, "IsBusy after idle");
      });
    }

    [Test]
    public void TestStopWaitForMessageArrivalWatchOperationNotStarted()
    {
      TestUtils.TestOpenedMailbox(new[] {ImapCapability.Idle},
                                  "INBOX",
                                  "$tag OK [READ-WRITE] done\r\n",
                                  delegate(ImapPseudoServer server, ImapOpenedMailboxInfo mailbox) {
        Assert.Throws<ImapProtocolViolationException>(() => mailbox.StopWaitForMessageArrival());
      });
    }

    [Test]
    public void TestStopWaitForMessageArrivalAnotherWatchOperationProceeding()
    {
      TestUtils.TestOpenedMailbox(new[] {ImapCapability.Idle},
                                  "INBOX",
                                  "$tag OK [READ-WRITE] done\r\n",
                                  delegate(ImapPseudoServer server, ImapOpenedMailboxInfo mailbox) {
        server.EnqueueResponse("+ idling\r\n");

        mailbox.StartIdle();

        Thread.Sleep(25); // XXX: wait for idle started

        Assert.Throws<ImapProtocolViolationException>(() => mailbox.StopWaitForMessageArrival());

        server.EnqueueTaggedResponse("$tag OK done\r\n");

        mailbox.StopIdle();
      });
    }

    [Test]
    public void TestWatchOperationProceedingRefresh()
    {
      TestUtils.TestOpenedMailbox(new[] {ImapCapability.Idle},
                                  "INBOX",
                                  "$tag OK [READ-WRITE] done\r\n",
                                  delegate(ImapPseudoServer server, ImapOpenedMailboxInfo mailbox) {
        Assert.IsFalse(mailbox.Client.IsBusy, "IsBusy before idle");

        server.EnqueueResponse("+ idling\r\n");

        mailbox.StartIdle();

        Thread.Sleep(25); // XXX: wait for idle started

        Assert.IsTrue(mailbox.Client.IsBusy, "IsBusy while idling");

        Assert.Throws<InvalidOperationException>(() => mailbox.Refresh());

        server.EnqueueTaggedResponse("$tag OK done\r\n");

        mailbox.StopIdle();

        Assert.IsFalse(mailbox.Client.IsBusy, "IsBusy after idle");
      });
    }

    [Test]
    public void TestWatchOperationProceedingClose()
    {
      TestUtils.TestOpenedMailbox(new[] {ImapCapability.Idle},
                                  "INBOX",
                                  "$tag OK [READ-WRITE] done\r\n",
                                  delegate(ImapPseudoServer server, ImapOpenedMailboxInfo mailbox) {
        Assert.IsFalse(mailbox.Client.IsBusy, "IsBusy before idle");

        server.EnqueueResponse("+ idling\r\n");

        mailbox.StartIdle();

        Thread.Sleep(25); // XXX: wait for idle started

        Assert.IsTrue(mailbox.Client.IsBusy, "IsBusy while idling");

        Assert.Throws<InvalidOperationException>(() => mailbox.Close());

        server.EnqueueTaggedResponse("$tag OK done\r\n");

        mailbox.StopIdle();

        Assert.IsFalse(mailbox.Client.IsBusy, "IsBusy after idle");
      });
    }
  }
}

