// 
// Author:
//       smdn <smdn@smdn.jp>
// 
// Copyright (c) 2008-2014 smdn
// 
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
// 
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
// 
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
// THE SOFTWARE.

using System;
using System.Collections.Generic;
using System.Net.Mail;
using System.Text;

using Smdn.Formats;

namespace Smdn.Net.Imap4 {
  // 7.4.2. FETCH Response
  //       ENVELOPE
  //         An address structure is a parenthesized list that describes an
  //         electronic mail address.  The fields of an address structure
  //         are in the following order: personal name, [SMTP]
  //         at-domain-list (source route), mailbox name, and host name.
  public sealed class ImapAddress {
    public string Name {
      get { return name; }
    }

    public string AtDomailList {
      get { return atDomainList; }
    }

    public string Mailbox {
      get { return mailbox; }
    }

    public string Host {
      get { return host; }
    }

    public ImapAddress(string name, string atDomainList, string mailbox, string host)
    {
      this.name = name;
      this.atDomainList = atDomainList;
      this.mailbox = mailbox;
      this.host = host;
    }

    public string GetDecodedName()
    {
      Encoding discard;

      return GetDecodedName(null, false, out discard);
    }

    public string GetDecodedName(EncodingSelectionCallback selectFallbackEncoding,
                                 bool throwIfMalformedOrUnsupported)
    {
      Encoding discard;

      return GetDecodedName(selectFallbackEncoding,
                            throwIfMalformedOrUnsupported,
                            out discard);
    }

    private string GetDecodedName(EncodingSelectionCallback selectFallbackEncoding,
                                  bool throwIfMalformedOrUnsupported,
                                  out Encoding displayNameEncoding)
    {
      return MimeUtils.GetDecodedValue(name,
                                       selectFallbackEncoding,
                                       throwIfMalformedOrUnsupported,
                                       out displayNameEncoding);
    }

    public MailAddress ToMailAddress()
    {
      return ToMailAddress(null, false);
    }

    public MailAddress ToMailAddress(EncodingSelectionCallback selectFallbackEncoding,
                                     bool throwIfMalformedOrUnsupported)
    {
      Encoding displayNameEncoding;
      var address = string.Concat(mailbox, "@", host);
      var decodedDisplayName = GetDecodedName(selectFallbackEncoding,
                                              throwIfMalformedOrUnsupported,
                                              out displayNameEncoding);

      if (decodedDisplayName == null)
        return new MailAddress(address);
      else
        return new MailAddress(address, decodedDisplayName, displayNameEncoding);
    }

    public static MailAddressCollection ToMailAddressCollection(IEnumerable<ImapAddress> addresses)
    {
      return ToMailAddressCollection(addresses, null, false);
    }

    public static MailAddressCollection ToMailAddressCollection(IEnumerable<ImapAddress> addresses,
                                                                EncodingSelectionCallback selectFallbackEncoding,
                                                                bool throwIfMalformedOrUnsupported)
    {
      if (addresses == null)
        throw new ArgumentNullException("addresses");

      var addressCollection = new MailAddressCollection();

      foreach (var address in addresses) {
        addressCollection.Add(address.ToMailAddress(selectFallbackEncoding, throwIfMalformedOrUnsupported));
      }

      return addressCollection;
    }

    public override string ToString()
    {
      return string.Format("{{Name={0}, AtDomainList={1}, Mailbox={2}, Host={3}}}", name, atDomainList, mailbox, host);
    }

    private /*readonly*/ string name;
    private /*readonly*/ string atDomainList;
    private /*readonly*/ string mailbox;
    private /*readonly*/ string host;
  }
}

/*
address         = "(" addr-name SP addr-adl SP addr-mailbox SP
                  addr-host ")"
addr-adl        = nstring
                    ; Holds route from [RFC-2822] route-addr if
                    ; non-NIL
addr-host       = nstring
                    ; NIL indicates [RFC-2822] group syntax.
                    ; Otherwise, holds [RFC-2822] domain name
addr-mailbox    = nstring
                    ; NIL indicates end of [RFC-2822] group; if
                    ; non-NIL and addr-host is NIL, holds
                    ; [RFC-2822] group name.
                    ; Otherwise, holds [RFC-2822] local-part
                    ; after removing [RFC-2822] quoting
addr-name       = nstring
                    ; If non-NIL, holds phrase from [RFC-2822]
                    ; mailbox after removing [RFC-2822] quoting
 */