// 
// Author:
//       smdn <smdn@smdn.jp>
// 
// Copyright (c) 2008-2014 smdn
// 
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
// 
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
// 
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
// THE SOFTWARE.

using System;
using System.Collections;
using System.Collections.Generic;
using System.Linq;
using System.Runtime.Serialization;

namespace Smdn.Net.Imap4 {
  // seq-number      = nz-number / "*"
  // seq-range       = seq-number ":" seq-number
  // sequence-set    = (seq-number / seq-range) *("," sequence-set)
  [Serializable]
  public abstract class ImapSequenceSet : IEnumerable<long>, ISerializable {
    public virtual bool IsUidSet {
      get; private set;
    }

    public virtual bool IsEmpty {
      get { return false; }
    }

    public abstract bool IsSingle { get; }
    /*
    public abstract long PossibleMin { get; }
    public abstract long PossibleMax { get; }
    */

    public abstract long ConcreteCount { get; }

    protected ImapSequenceSet(bool isUidSet)
    {
      this.IsUidSet = isUidSet;
    }

    protected ImapSequenceSet(SerializationInfo info, StreamingContext context)
    {
      IsUidSet = info.GetBoolean("IsUidSet");
    }

    public virtual void GetObjectData(SerializationInfo info, StreamingContext context)
    {
      info.AddValue("IsUidSet", IsUidSet);
    }

    public virtual long ToNumber()
    {
      if (IsSingle)
        return ToArray()[0];
      else
        throw new NotSupportedException("can't get single number from non-single set");
    }

    public abstract long[] ToArray();
    public abstract IEnumerator<long> GetEnumerator();
    public override abstract string ToString();

    IEnumerator IEnumerable.GetEnumerator()
    {
      return GetEnumerator();
    }

    public virtual IEnumerable<ImapSequenceSet> SplitIntoEach(int count)
    {
      if (count <= 0)
        throw ExceptionUtils.CreateArgumentMustBeNonZeroPositive("count", count);

      var numbers = new long[count];
      var index = 0;
      var sequential = true;

      foreach (var number in this) {
        if (0 < index)
          sequential &= (numbers[index - 1] == number - 1L);

        numbers[index++] = number;

        if (count <= index) {
          if (sequential) {
            yield return CreateRangeSet(IsUidSet, numbers[0], number);
          }
          else {
            yield return CreateSetUncheckedDirect(IsUidSet, numbers);

            numbers = new long[count];
          }

          index = 0;
          sequential = true;
        }
      }

      if (0 < index) {
        if (sequential)
          yield return CreateRangeSet(IsUidSet, numbers[0], numbers[index - 1]);
        else
          yield return CreateSet(IsUidSet, numbers, 0, index);
      }
    }

    public static ImapSequenceSet FromUri(Uri uri)
    {
      var uid = ImapStyleUriParser.GetUid(uri);

      if (uid == 0L)
        throw new ArgumentException("invalid IMAP-URI form; URI has no UID", "uri");

      return CreateUidSet(uid);
    }

#region "create from:* set"
    public static ImapSequenceSet CreateFromSet(long from)
    {
      return CreateFromSet(false, from);
    }

    public static ImapSequenceSet CreateUidFromSet(long from)
    {
      return CreateFromSet(true, from);
    }

    public static ImapSequenceSet CreateFromSet(bool uidSet, long from)
    {
      return new FromSequenceSet(uidSet, from);
    }
#endregion

#region "create *:to set"
    public static ImapSequenceSet CreateToSet(long to)
    {
      return CreateToSet(false, to);
    }

    public static ImapSequenceSet CreateUidToSet(long to)
    {
      return CreateToSet(true, to);
    }

    public static ImapSequenceSet CreateToSet(bool uidSet, long to)
    {
      return new ToSequenceSet(uidSet, to);
    }
#endregion

#region "create from:to set"
    public static ImapSequenceSet CreateRangeSet(long from, long to)
    {
      return CreateRangeSet(false, from, to);
    }

    public static ImapSequenceSet CreateUidRangeSet(long from, long to)
    {
      return CreateRangeSet(true, from, to);
    }

    public static ImapSequenceSet CreateRangeSet(bool uidSet, long from, long to)
    {
      if (from == to)
        return new DiscreteSequenceSet(uidSet, new[] {from});
      else
        return new RangeSequenceSet(uidSet, from, to);
    }
#endregion

#region "create * set"
    public static ImapSequenceSet CreateAllSet()
    {
      return CreateAllSet(false);
    }

    public static ImapSequenceSet CreateUidAllSet()
    {
      return CreateAllSet(true);
    }

    public static ImapSequenceSet CreateAllSet(bool uid)
    {
      return new AllSequenceSet(uid);
    }
#endregion

#region "create num*(,num) set"
    public static ImapSequenceSet CreateSet(long number, params long[] numbers)
    {
      return CreateSetUncheckedDirect(false, numbers.Prepend(number));
    }

    public static ImapSequenceSet CreateUidSet(long number, params long[] numbers)
    {
      return CreateSetUncheckedDirect(true, numbers.Prepend(number));
    }

    public static ImapSequenceSet CreateSet(bool uidSet, long number, params long[] numbers)
    {
      return CreateSetUncheckedDirect(uidSet, numbers.Prepend(number));
    }

    public static ImapSequenceSet CreateSet(long[] numbers)
    {
      return CreateSet(false, numbers);
    }

    public static ImapSequenceSet CreateUidSet(long[] numbers)
    {
      return CreateSet(true, numbers);
    }

    public static ImapSequenceSet CreateSet(bool uidSet, long[] numbers)
    {
      if (numbers == null)
        throw new ArgumentNullException("numbers");

      return new DiscreteSequenceSet(uidSet, (long[])numbers.Clone());
    }

    internal static ImapSequenceSet CreateSetUncheckedDirect(bool uidSet, long[] numbers)
    {
      return new DiscreteSequenceSet(uidSet, numbers);
    }

    public static ImapSequenceSet CreateSet(long[] numbers, int start, int count)
    {
      return CreateSet(false, numbers, start, count);
    }

    public static ImapSequenceSet CreateUidSet(long[] numbers, int start, int count)
    {
      return CreateSet(true, numbers, start, count);
    }

    public static ImapSequenceSet CreateSet(bool uidSet, long[] numbers, int start, int count)
    {
      if (numbers == null)
        throw new ArgumentNullException("numbers");
      if (count <= 0)
        throw ExceptionUtils.CreateArgumentMustBeNonZeroPositive("count", count);
      if (start < 0)
        throw ExceptionUtils.CreateArgumentMustBeZeroOrPositive("start", start);
      if (numbers.Length - count < start)
        throw ExceptionUtils.CreateArgumentAttemptToAccessBeyondEndOfArray("start", numbers, start, count);

      return new DiscreteSequenceSet(uidSet, numbers.Slice(start, count));
    }
#endregion

#region "create combined set"
    public static ImapSequenceSet operator+ (ImapSequenceSet x, ImapSequenceSet y)
    {
      return Combine(x, y);
    }

    public ImapSequenceSet CombineWith(ImapSequenceSet other)
    {
      return Combine(this, other);
    }

    public static ImapSequenceSet Combine(ImapSequenceSet x, ImapSequenceSet y)
    {
      if (x == null)
        return y;
      else if (y == null)
        return x;

      if (x.IsUidSet != y.IsUidSet)
        throw new ArgumentException("can't contain simultaneously both UID set and sequence set", "x");

      var xx = x as ImapMatchedSequenceSet;
      var yy = y as ImapMatchedSequenceSet;

      if (xx != null) {
        if (xx.IsSavedResult)
          throw new ArgumentException("can't combine with saved result", "x");
        x = xx.SequenceSet;
      }

      if (yy != null) {
        if (yy.IsSavedResult)
          throw new ArgumentException("can't combine with saved result", "y");
        y = yy.SequenceSet;
      }

      var cx = x as CombinedSequenceSet;
      var cy = y as CombinedSequenceSet;
      var capacity = (cx == null ? 1 : cx.Sets.Count) + (cy == null ? 1 : cy.Sets.Count);
      var allSets = new List<ImapSequenceSet>(capacity);

      if (cx == null)
        allSets.Add(x);
      else
        allSets.AddRange(cx.Sets);

      if (cy == null)
        allSets.Add(y);
      else
        allSets.AddRange(cy.Sets);

      return GetRegularizedSet(x.IsUidSet, allSets, false);
    }

    internal static ImapSequenceSet GetRegularizedSet(bool uid, ICollection<ImapSequenceSet> sets, bool regularizeRangedSet)
    {
      if (regularizeRangedSet)
        throw new NotImplementedException();

      var filteredSets = new List<ImapSequenceSet>(sets.Count);
#if REGULARIZE_RANGED_SET
      var fromMin = long.MaxValue;
      var toMax = 0L;
#endif

      foreach (var s in sets) {
        if (s is AllSequenceSet)
          return new AllSequenceSet(uid);

        if (s.IsEmpty)
          continue;

#if REGULARIZE_RANGED_SET
        var fromSet = s as FromSequenceSet;

        if (fromSet != null && fromSet.From < fromMin) {
          fromMin = fromSet.From;
          continue;
        }

        var toSet = s as ToSequenceSet;

        if (toSet != null && toMax < toSet.To) {
          toMax = toSet.To;
          continue;
        }
#endif

        filteredSets.Add(s);
      }

      // TODO: regularize 'from:to' and 'n1,n2,n3...'
#if REGULARIZE_RANGED_SET
      if (fromMin != long.MaxValue)
        filteredSets.Insert(0, ImapSequenceSet.CreateFromSet(uid, fromMin));
      if (toMax != 0)
        filteredSets.Add(ImapSequenceSet.CreateToSet(uid, toMax));
#endif

      return new CombinedSequenceSet(uid, filteredSets);
    }
#endregion

#region "sequence set classes"
    private interface IFromSequenceSet {
      long From { get; }
    }

    private interface IToSequenceSet {
      long To { get; }
    }

    // seq-number:*
    [Serializable]
    private class FromSequenceSet : ImapSequenceSet, IFromSequenceSet {
      public long From {
        get; private set;
      }

      public override bool IsSingle {
        get { return false; }
      }

      /*
      public override long PossibleMax {
        get { return long.MaxValue; }
      }

      public override long PossibleMin {
        get { return From; }
      }
      */

      public override long ConcreteCount {
        get { return -1; }
      }

      public FromSequenceSet(bool isUidSet, long from)
        : base(isUidSet)
      {
        if (from <= 0)
          throw ExceptionUtils.CreateArgumentMustBeNonZeroPositive("from", from);

        this.From = from;
      }

      protected FromSequenceSet(SerializationInfo info, StreamingContext context)
        : base(info, context)
      {
        From = info.GetInt64("From");
      }

      public override void GetObjectData(SerializationInfo info, StreamingContext context)
      {
        base.GetObjectData(info, context);

        info.AddValue("From", From);
      }

      public override IEnumerator<long> GetEnumerator()
      {
        throw new NotSupportedException("can't enumerate 'n:*' set");
      }

      public override long[] ToArray()
      {
        throw new NotSupportedException("can't create array from 'n:*' set");
      }

      public override string ToString()
      {
        return string.Concat(From, ":*");
      }
    }

    // *:seq-number
    [Serializable]
    private class ToSequenceSet : ImapSequenceSet, IToSequenceSet {
      public long To {
        get; private set;
      }

      public override bool IsSingle {
        get { return false; }
      }

      /*
      public override long PossibleMax {
        get { return To; }
      }

      public override long PossibleMin {
        get { return 1L; }
      }
      */

      public override long ConcreteCount {
        get { return To; }
      }

      public ToSequenceSet(bool isUidSet, long to)
        : base(isUidSet)
      {
        if (to <= 0)
          throw ExceptionUtils.CreateArgumentMustBeNonZeroPositive("to", to);

        this.To = to;
      }

      protected ToSequenceSet(SerializationInfo info, StreamingContext context)
        : base(info, context)
      {
        To = info.GetInt64("To");
      }

      public override void GetObjectData(SerializationInfo info, StreamingContext context)
      {
        base.GetObjectData(info, context);

        info.AddValue("To", To);
      }

      public override IEnumerator<long> GetEnumerator()
      {
        throw new NotSupportedException("can't enumerate '*:n' set");
      }

      public override long[] ToArray()
      {
        throw new NotSupportedException("can't create array from '*:n' set");
      }

      public override string ToString()
      {
        return string.Concat("*:", To);
      }
    }

    // seq-number:seq-number
    [Serializable]
    private class RangeSequenceSet : ImapSequenceSet, IFromSequenceSet, IToSequenceSet {
      public long From {
        get; private set;
      }

      public long To {
        get; private set;
      }

      public override bool IsSingle {
        get { return false /*From == To*/; }
      }

      /*
      public override long PossibleMax {
        get { return To; }
      }

      public override long PossibleMin {
        get { return From; }
      }
      */

      public override long ConcreteCount {
        get { return (To - From) + 1; }
      }

      public RangeSequenceSet(bool isUidSet, long from, long to)
        : base(isUidSet)
      {
        if (from <= 0)
          throw ExceptionUtils.CreateArgumentMustBeNonZeroPositive("from", from);
        if (to <= from)
          throw ExceptionUtils.CreateArgumentMustBeGreaterThan("'from'", "to", to);

        this.From = from;
        this.To = to;
      }

      protected RangeSequenceSet(SerializationInfo info, StreamingContext context)
        : base(info, context)
      {
        From = info.GetInt64("From");
        To = info.GetInt64("To");
      }

      public override void GetObjectData(SerializationInfo info, StreamingContext context)
      {
        base.GetObjectData(info, context);

        info.AddValue("From", From);
        info.AddValue("To", To);
      }

      public override IEnumerator<long> GetEnumerator()
      {
        for (var i = From; i <= To; i++) {
          yield return i;
        }
      }

      public override long[] ToArray()
      {
        var arr = new long[To - From + 1];
        var index = 0;

        for (var i = From; i <= To; i++) {
          arr[index++] = i;
        }

        return arr;
      }

      public override string ToString()
      {
        return string.Concat(From, ":", To);
      }
    }

    // *
    [Serializable]
    private class AllSequenceSet : ImapSequenceSet {
      public override bool IsSingle {
        get { return false; }
      }

      /*
      public override long PossibleMax {
        get { return long.MaxValue; }
      }

      public override long PossibleMin {
        get { return 1L; }
      }
      */

      public override long ConcreteCount {
        get { return -1; }
      }

      public AllSequenceSet(bool isUidSet)
        : base(isUidSet)
      {
      }

      protected AllSequenceSet(SerializationInfo info, StreamingContext context)
        : base(info, context)
      {
      }

      /*
      public override void GetObjectData(SerializationInfo info, StreamingContext context)
      {
        base.GetObjectData(info, context);
      }
      */

      public override IEnumerator<long> GetEnumerator()
      {
        throw new NotSupportedException("can't enumerate '*' set");
      }

      public override long[] ToArray()
      {
        throw new NotSupportedException("can't create array from '*' set");
      }

      public override string ToString()
      {
        return "*";
      }
    }

    // seq-number *("," sequence-set)
    [Serializable]
    private class DiscreteSequenceSet : ImapSequenceSet {
      public override bool IsEmpty {
        get { return Set.Length == 0; }
      }

      public long[] Set {
        get; private set;
      }

      public override bool IsSingle {
        get { return Set.Length == 1; }
      }

      /*
      public override long PossibleMax {
        get { return Set[Set.Length - 1]; // must be sorted }
      }

      public override long PossibleMin {
        get { return Set[0]; // must be sorted }
      }
      */

      public override long ConcreteCount {
        get { return Set.Length; }
      }

      public DiscreteSequenceSet(bool isUidSet, long[] @set)
        : base(isUidSet)
      {
        if (@set == null)
          throw new ArgumentNullException("set");

        for (var i = 0; i < @set.Length; i++) {
          if (@set[i] <= 0)
            throw new ArgumentOutOfRangeException("sets",
                                                  @set[i],
                                                  string.Format("must be non-zero positive number (index: {0})", i));
        }

        this.Set = @set;
      }

      protected DiscreteSequenceSet(SerializationInfo info, StreamingContext context)
        : base(info, context)
      {
        Set = (long[])info.GetValue("Set", typeof(long[]));
      }

      public override void GetObjectData(SerializationInfo info, StreamingContext context)
      {
        base.GetObjectData(info, context);

        info.AddValue("Set", Set, typeof(long[]));
      }

      public override long ToNumber()
      {
        if (Set.Length == 0)
          throw new NotSupportedException("can't get single number from empty set");
        else if (Set.Length == 1)
          return Set[0];
        else
          throw new NotSupportedException("can't get single number from non-single set");
      }

      public override IEnumerator<long> GetEnumerator()
      {
        foreach (var num in Set) {
          yield return num;
        }
      }

      public override long[] ToArray()
      {
        return Set.Clone() as long[];
      }

      public override string ToString()
      {
        return string.Join(",", Array.ConvertAll(Set, delegate(long val) {
          return val.ToString();
        }));
      }
    }

    // (seq-number / seq-range) *("," sequence-set)
    [Serializable]
    private class CombinedSequenceSet : ImapSequenceSet {
      private static bool IsSingleSet(ImapSequenceSet s)
      {
        return s.IsSingle;
      }

      public ICollection<ImapSequenceSet> Sets {
        get; private set;
      }

      public override bool IsSingle {
        get { return Sets.All(IsSingleSet); }
      }

      /*
      public override long PossibleMax {
        get
        {
          var max = 1L;

          foreach (var @set in Sets) {
            if (max < @set.PossibleMax)
              max = @set.PossibleMax;
          }

          return max;
        }
      }

      public override long PossibleMin {
        get
        {
          var min = long.MaxValue;

          foreach (var @set in Sets) {
            if (@set.PossibleMin < min)
              min = @set.PossibleMin;
          }

          return min;
        }
      }
      */

      public override long ConcreteCount {
        get {
          if (Sets.Count == 0)
            return -1;

          var c = 0L;

          foreach (var @set in Sets) {
            if (@set.ConcreteCount == -1)
              return -1;
            else
              c += @set.ConcreteCount;
          }

          return c;
        }
      }

      public CombinedSequenceSet(bool isUidSet, ICollection<ImapSequenceSet> sets)
        : base(isUidSet)
      {
        if (sets == null)
          throw new ArgumentNullException("sets");

        foreach (var s in sets) {
          if (s.IsUidSet != isUidSet)
            throw new ArgumentException("can't contain simultaneously both UID set and sequence set", "sets");
        }

        this.Sets = sets;
      }

      protected CombinedSequenceSet(SerializationInfo info, StreamingContext context)
        : base(info, context)
      {
        Sets = (ImapSequenceSet[])info.GetValue("Sets", typeof(ImapSequenceSet[]));
      }

      public override void GetObjectData(SerializationInfo info, StreamingContext context)
      {
        base.GetObjectData(info, context);

        info.AddValue("Sets", Sets.ToArray(), typeof(ImapSequenceSet[]));
      }

      public override long ToNumber()
      {
        var single = Sets.FirstOrDefault(IsSingleSet);

        if (single == null)
          throw new NotSupportedException("can't get single number from non-single set");
        else
          return single.ToNumber();
      }

      public override IEnumerator<long> GetEnumerator()
      {
        foreach (var s in Sets) {
          foreach (var num in s) {
            yield return num;
          }
        }
      }

      public override long[] ToArray()
      {
        var arr = new List<long>();

        foreach (var s in Sets) {
          arr.AddRange(s.ToArray());
        }

        return arr.ToArray();
      }

      public override string ToString()
      {
        return string.Join(",", Sets.Select(delegate(ImapSequenceSet @set) {
          return @set.ToString();
        })
#if !NET_4_0
        .ToArray()
#endif  
        );
      }
    }
#endregion
  }
}