using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using NUnit.Framework;

using Smdn.Formats;

namespace Smdn.Net.Imap4 {
  [TestFixture]
  public class ImapBodyStructureUtilsTests {
    private IImapBodyStructure BuildMultiPart()
    {
      var part1 = new ImapSinglePartBodyStructure("1", new MimeType("text", "plain"), null, null, null, "7bit", 0L, 0L);

      var part21 = new ImapSinglePartBodyStructure("2.1", new MimeType("text", "plain"), null, null, null, "7bit", 0L, 0L);
      var part22 = new ImapSinglePartBodyStructure("2.2", new MimeType("text", "html"), null, null, null, "7bit", 0L, 0L);
      var part2 = new ImapMultiPartBodyStructure("2",
                                                 new[] {part21, part22},
                                                 "alternative");

      var part31 = new ImapSinglePartBodyStructure("3.1", new MimeType("text", "plain"), null, null, null, "7bit", 0L, 0L);
      var part3Envelope = new ImapEnvelope(null, null, null, null, null, null, null, null, null, null);
      var part3 = new ImapMessageRfc822BodyStructure("3", new MimeType("message", "rfc822"), null, null, null, "7bit", 0L, part3Envelope, part31, 0L);

      return new ImapMultiPartBodyStructure(string.Empty,
                                            new IImapBodyStructure[] {part1, part2, part3},
                                            "mixed");
    }

    [Test]
    public void TestTraverseSinglePart()
    {
      var single = new ImapSinglePartBodyStructure(string.Empty, new MimeType("text", "plain"), null, null, null, "7bit", 0L, 0L);

      ImapBodyStructureUtils.Traverse(single, delegate(IImapBodyStructure s) {
        Assert.Fail("enumerated");
      });
    }

    [Test]
    public void TestTraverseMultiPart()
    {
      var expectedSections = new[] {"1", "2", "2.1", "2.2", "3", "3.1"};
      var index = 0;

      ImapBodyStructureUtils.Traverse(BuildMultiPart(), delegate(IImapBodyStructure s) {
        Assert.AreEqual(expectedSections[index++], s.Section);
      });
    }

    [Test]
    public void TestFindSection1()
    {
      var structure = BuildMultiPart();
      var found = structure.FindSection(3, 1);

      Assert.IsNotNull(found);
      Assert.AreEqual("3.1", found.Section);
      Assert.AreEqual(MimeType.TextPlain, found.MediaType);

      found = structure.FindSection(4, 1);

      Assert.IsNull(found);

      Assert.IsNull(structure.FindSection());
    }

    [Test]
    public void TestFindSection2()
    {
      var structure = BuildMultiPart();
      var found = structure.FindSection("3");

      Assert.IsNotNull(found);
      Assert.AreEqual("3", found.Section);
      Assert.AreEqual(new MimeType("message/rfc822"), found.MediaType);

      found = structure.FindSection("4.1");

      Assert.IsNull(found);

      Assert.IsNull(structure.FindSection((string)null));
      Assert.IsNull(structure.FindSection(string.Empty));
    }

    [Test]
    public void TestFind()
    {
      var structure = BuildMultiPart();
      var found = structure.Find(delegate(IImapBodyStructure s) {
        return s is ImapMessageRfc822BodyStructure;
      });

      Assert.IsNotNull(found);
      Assert.AreEqual("3", found.Section);

      found = structure.Find(delegate(IImapBodyStructure s) {
        return s is ImapExtendedMessageRfc822BodyStructure;
      });

      Assert.IsNull(found);
    }

    [Test]
    public void TestFindAll()
    {
      var structure = BuildMultiPart();
      var found = structure.FindAll(delegate(IImapBodyStructure s) {
        return !s.IsMultiPart;
      });

      var expectedSections = new[] {"1", "2.1", "2.2", "3", "3.1"};
      var index = 0;

      Assert.IsNotNull(found);

      foreach (var s in found) {
        Assert.IsNotNull(s);
        Assert.AreEqual(expectedSections[index++], s.Section);
      }

      found = structure.FindAll(delegate(IImapBodyStructure s) {
        return s is IImapBodyStructureExtension;
      });

      Assert.IsNotNull(found);
      Assert.AreEqual(0, found.Count());
    }

    [Test]
    public void TestFindByMediaType()
    {
      var structure = BuildMultiPart();
      var found = structure.Find(new MimeType("MESSAGE/RFC822"));

      Assert.IsNotNull(found);
      Assert.AreEqual("3", found.Section);

      found = structure.Find(MimeType.ApplicationOctetStream);

      Assert.IsNull(found);
    }

    [Test]
    public void TestFindAllByMediaType()
    {
      var structure = BuildMultiPart();
      var found = structure.FindAll(MimeType.TextPlain);

      var expectedSections = new[] {"1", "2.1", "3.1"};
      var index = 0;

      Assert.IsNotNull(found);

      foreach (var s in found) {
        Assert.IsNotNull(s);
        Assert.AreEqual(expectedSections[index++], s.Section);
      }

      found = structure.FindAll(MimeType.ApplicationOctetStream);

      Assert.IsNotNull(found);
      Assert.AreEqual(0, found.Count());
    }

    [Test]
    public void TestGetRootStructure()
    {
      var structure = BuildMultiPart();

      Assert.AreSame(structure, structure.GetRootStructure());
      Assert.AreSame(structure, structure.FindSection("1").GetRootStructure());
      Assert.AreSame(structure, structure.FindSection("2").GetRootStructure());
      Assert.AreSame(structure, structure.FindSection("2.1").GetRootStructure());
      Assert.AreSame(structure, structure.FindSection("2.2").GetRootStructure());
      Assert.AreSame(structure, structure.FindSection("3").GetRootStructure());
      Assert.AreSame(structure, structure.FindSection("3.1").GetRootStructure());
    }

    [Test]
    public void TestGetContentFileNameFromContentTypeNameParameter()
    {
      var structure1 =
        new ImapSinglePartBodyStructure("1",
                                        new MimeType("image", "jpeg"),
                                        new Dictionary<string, string>() {{"name", "attached.jpg"}},
                                        null,
                                        null,
                                        "7bit",
                                        0L,
                                        0L);

      Assert.IsTrue(structure1.Parameters.ContainsKey("name"));
      Assert.AreEqual("attached.jpg", structure1.Parameters["name"]);
      Assert.AreEqual("attached.jpg", structure1.GetContentFileName());

      var structure2 =
        new ImapSinglePartBodyStructure("1",
                                        new MimeType("image", "jpeg"),
                                        new Dictionary<string, string>() {{"name", "=?ISO-2022-JP?B?GyRCRTpJVSVVJSElJCVrGyhCLmpwZw==?="}},
                                        null,
                                        null,
                                        "7bit",
                                        0L,
                                        0L);

      Assert.IsTrue(structure2.Parameters.ContainsKey("name"));
      Assert.AreEqual("=?ISO-2022-JP?B?GyRCRTpJVSVVJSElJCVrGyhCLmpwZw==?=", structure2.Parameters["name"]);
      Assert.AreEqual("添付ファイル.jpg", structure2.GetContentFileName());
    }

    [Test]
    public void TestGetContentFileNameFromContentDispositionFilenameParameter()
    {
      var baseStructure =
        new ImapSinglePartBodyStructure("1",
                                        new MimeType("image", "jpeg"),
                                        null,
                                        null,
                                        null,
                                        "7bit",
                                        0L,
                                        0L);

      var structure =
        new ImapExtendedSinglePartBodyStructure(baseStructure,
                                                null,
                                                new ImapBodyDisposition("attached",
                                                                        new Dictionary<string, string>() {{"filename", "attached.jpg"}}),
                                                null,
                                                null,
                                                null);

      Assert.IsTrue(structure.Disposition.Parameters.ContainsKey("filename"));
      Assert.AreEqual("attached.jpg", structure.Disposition.Parameters["filename"]);
      Assert.IsFalse(structure.Disposition.Parameters.ContainsKey("filename*"));
      Assert.AreEqual("attached.jpg", structure.GetContentFileName());
    }

    [Test]
    public void TestGetContentFileNameFromContentDispositionExtendedFilenameParameter()
    {
      var baseStructure =
        new ImapSinglePartBodyStructure("1",
                                        new MimeType("image", "jpeg"),
                                        null,
                                        null,
                                        null,
                                        "7bit",
                                        0L,
                                        0L);

      var structure =
        new ImapExtendedSinglePartBodyStructure(baseStructure,
                                                null,
                                                new ImapBodyDisposition("attached",
                                                                        new Dictionary<string, string>() {{"filename*", "ISO-2022-JP''%1B%24%42%45%3A%49%55%25%55%25%21%25%24%25%6B%1B%28%42%2E%6A%70%67"}}),
                                                null,
                                                null,
                                                null);

      Assert.IsFalse(structure.Disposition.Parameters.ContainsKey("filename"));
      Assert.IsTrue(structure.Disposition.Parameters.ContainsKey("filename*"));
      Assert.AreEqual("ISO-2022-JP''%1B%24%42%45%3A%49%55%25%55%25%21%25%24%25%6B%1B%28%42%2E%6A%70%67", structure.Disposition.Parameters["filename*"]);
      Assert.AreEqual("添付ファイル.jpg", structure.GetContentFileName());
    }

    [Test]
    public void TestGetContentFileNameNoNameOrDisposition()
    {
      var baseStructure =
        new ImapSinglePartBodyStructure("1",
                                        new MimeType("image", "jpeg"),
                                        null,
                                        null,
                                        null,
                                        "7bit",
                                        0L,
                                        0L);

      var structure =
        new ImapExtendedSinglePartBodyStructure(baseStructure,
                                                null,
                                                null,
                                                null,
                                                null,
                                                null);

      Assert.IsFalse(structure.Parameters.ContainsKey("name"));
      Assert.IsNull(structure.Disposition);

      Assert.IsNull(structure.GetContentFileName());
    }

    [Test]
    public void TestGetContentFileNameWithExtendedMultiPartBodyStructure()
    {
      var nestedStructures = new[] {
        new ImapSinglePartBodyStructure("1.1",
                                        new MimeType("image", "jpeg"),
                                        null,
                                        null,
                                        null,
                                        null,
                                        0L,
                                        0L),
      };
      var baseStructure =
        new ImapMultiPartBodyStructure("1",
                                       nestedStructures,
                                       "mixed");

      var structure =
        new ImapExtendedMultiPartBodyStructure(baseStructure,
                                               null,
                                               new ImapBodyDisposition("attached",
                                                                       new Dictionary<string, string>() {{"filename", "attached.eml"}}),
                                               null,
                                               null,
                                               null);

      Assert.IsTrue(structure.Disposition.Parameters.ContainsKey("filename"));
      Assert.AreEqual("attached.eml", structure.Disposition.Parameters["filename"]);
      Assert.AreEqual("attached.eml", structure.GetContentFileName());
    }

    [Test]
    public void TestGetContentFileNameWithExtendedMessageRfc822BodyStructure()
    {
      var encapsulatedStructure =
        new ImapSinglePartBodyStructure("1.1",
                                        new MimeType("image", "jpeg"),
                                        null,
                                        null,
                                        null,
                                        null,
                                        0L,
                                        0L);

      var baseStructure =
        new ImapMessageRfc822BodyStructure("1",
                                           new MimeType("message", "rfc822"),
                                           null,
                                           null,
                                           null,
                                           null,
                                           0L,
                                           new ImapEnvelope(null, null, null, null, null, null, null, null, null, null),
                                           encapsulatedStructure,
                                           0L);

      var structure =
        new ImapExtendedMessageRfc822BodyStructure(baseStructure,
                                                   null,
                                                   new ImapBodyDisposition("attached",
                                                                       new Dictionary<string, string>() {{"filename", "attached.eml"}}),
                                                   null,
                                                   null,
                                                   null);

      Assert.IsTrue(structure.Disposition.Parameters.ContainsKey("filename"));
      Assert.AreEqual("attached.eml", structure.Disposition.Parameters["filename"]);
      Assert.AreEqual("attached.eml", structure.GetContentFileName());
    }

    [Test]
    public void TestGetContentFileNameNoNameOrFilenameParameter()
    {
      var baseStructure =
        new ImapSinglePartBodyStructure("1",
                                        new MimeType("image", "jpeg"),
                                        null,
                                        null,
                                        null,
                                        "7bit",
                                        0L,
                                        0L);

      var structure =
        new ImapExtendedSinglePartBodyStructure(baseStructure,
                                                null,
                                                new ImapBodyDisposition("attached", null),
                                                null,
                                                null,
                                                null);

      Assert.IsFalse(structure.Parameters.ContainsKey("name"));
      Assert.IsFalse(structure.Disposition.Parameters.ContainsKey("filename"));

      Assert.IsNull(structure.GetContentFileName());
    }

    [Test]
    public void TestGetContentFileNameNameParameterInvalidMimeEncoding()
    {
      var structure =
        new ImapSinglePartBodyStructure("1",
                                        new MimeType("image", "jpeg"),
                                        new Dictionary<string, string>() {{"name", "=?ISO-2022-JP?X?xxxxxx?="}},
                                        null,
                                        null,
                                        "7bit",
                                        0L,
                                        0L);

      Assert.AreEqual("=?ISO-2022-JP?X?xxxxxx?=", structure.GetContentFileName());
      Assert.AreEqual("=?ISO-2022-JP?X?xxxxxx?=", structure.GetContentFileName(null, false));

      Assert.Throws<FormatException>(() => structure.GetContentFileName(null, true));
    }

    [Test]
    public void TestGetContentFileNameNameParameterInvalidCharset()
    {
      var structure =
        new ImapSinglePartBodyStructure("1",
                                        new MimeType("image", "jpeg"),
                                        new Dictionary<string, string>() {{"name", "=?X-UNKNOWN?B?GyRCRTpJVSVVJSElJCVrGyhCLmpwZw==?="}},
                                        null,
                                        null,
                                        "7bit",
                                        0L,
                                        0L);

      Assert.AreEqual("=?X-UNKNOWN?B?GyRCRTpJVSVVJSElJCVrGyhCLmpwZw==?=", structure.GetContentFileName());
      Assert.AreEqual("=?X-UNKNOWN?B?GyRCRTpJVSVVJSElJCVrGyhCLmpwZw==?=", structure.GetContentFileName(null, false));

      Assert.Throws<EncodingNotSupportedException>(() => structure.GetContentFileName(null, true));

      Assert.AreEqual("添付ファイル.jpg", structure.GetContentFileName(delegate(string encoding) {
        Assert.AreEqual("X-UNKNOWN", encoding);

        return Encoding.GetEncoding("iso-2022-jp");
      }, true));
    }

    [Test]
    public void TestGetContentFileNameFilenameParameterInvalidCharset()
    {
      var baseStructure =
        new ImapSinglePartBodyStructure("1",
                                        new MimeType("image", "jpeg"),
                                        null,
                                        null,
                                        null,
                                        "7bit",
                                        0L,
                                        0L);

      var structure =
        new ImapExtendedSinglePartBodyStructure(baseStructure,
                                                null,
                                                new ImapBodyDisposition("attached",
                                                                        new Dictionary<string, string>() {{"filename*", "X-UNKNOWN''%1B%24%42%45%3A%49%55%25%55%25%21%25%24%25%6B%1B%28%42%2E%6A%70%67"}}),
                                                null,
                                                null,
                                                null);

      Assert.IsFalse(structure.Disposition.Parameters.ContainsKey("filename"));
      Assert.IsTrue(structure.Disposition.Parameters.ContainsKey("filename*"));
      Assert.AreEqual("X-UNKNOWN''%1B%24%42%45%3A%49%55%25%55%25%21%25%24%25%6B%1B%28%42%2E%6A%70%67", structure.Disposition.Parameters["filename*"]);
      Assert.AreEqual("X-UNKNOWN''%1B%24%42%45%3A%49%55%25%55%25%21%25%24%25%6B%1B%28%42%2E%6A%70%67", structure.GetContentFileName());
      Assert.AreEqual("X-UNKNOWN''%1B%24%42%45%3A%49%55%25%55%25%21%25%24%25%6B%1B%28%42%2E%6A%70%67", structure.GetContentFileName(null, false));

      Assert.Throws<EncodingNotSupportedException>(() => structure.GetContentFileName(null, true));

      Assert.AreEqual("添付ファイル.jpg", structure.GetContentFileName(delegate(string encoding) {
        Assert.AreEqual("X-UNKNOWN", encoding);

        return Encoding.GetEncoding("iso-2022-jp");
      }, true));
    }
  }
}
